from kopf._core.intents.piggybacking import has_service_account, login_with_service_account

# As per https://kubernetes.io/docs/tasks/run-application/access-api-from-pod/
NAMESPACE_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
SA_TOKEN_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/token'
CA_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt'


def test_has_no_serviceaccount_when_special_file_is_absent(mocker):
    exists_mock = mocker.patch('os.path.exists', return_value=False)
    result = has_service_account()
    assert result is False
    assert exists_mock.call_count == 1
    assert exists_mock.call_args_list[0][0][0] == SA_TOKEN_PATH


def test_has_serviceaccount_when_special_file_exists(mocker):
    exists_mock = mocker.patch('os.path.exists', return_value=True)
    result = has_service_account()
    assert result is True
    assert exists_mock.call_count == 1
    assert exists_mock.call_args_list[0][0][0] == SA_TOKEN_PATH


def test_serviceaccount_with_all_absent_files(mocker):
    exists_mock = mocker.patch('os.path.exists', return_value=False)  # all 3 of them.
    open_mock = mocker.patch('kopf._core.intents.piggybacking.open')
    open_mock.return_value.__enter__.return_value.read.return_value = ''
    credentials = login_with_service_account()
    assert credentials is None
    assert exists_mock.call_count == 1
    assert exists_mock.call_args_list[0][0][0] == SA_TOKEN_PATH
    assert not open_mock.called


def test_serviceaccount_with_all_present_files(mocker):
    exists_mock = mocker.patch('os.path.exists', return_value=True)  # all 3 of them.
    open_mock = mocker.patch('kopf._core.intents.piggybacking.open')
    open_mock.return_value.__enter__.return_value.read.side_effect=[' tkn ', ' ns ', RuntimeError]
    credentials = login_with_service_account()
    assert credentials is not None
    assert credentials.server == 'https://kubernetes.default.svc'
    assert credentials.default_namespace == 'ns'
    assert credentials.token == 'tkn'
    assert credentials.ca_path == CA_PATH
    assert exists_mock.call_count == 3
    assert exists_mock.call_args_list[0][0][0] == SA_TOKEN_PATH
    assert exists_mock.call_args_list[1][0][0] == NAMESPACE_PATH
    assert exists_mock.call_args_list[2][0][0] == CA_PATH
    assert open_mock.call_count == 2
    assert open_mock.call_args_list[0][0][0] == SA_TOKEN_PATH
    assert open_mock.call_args_list[1][0][0] == NAMESPACE_PATH
    # NB: the order is irrelevant and can be changed if needed.


def test_serviceaccount_with_only_the_token_file(mocker):
    # NB: the order is irrelevant and can be changed if needed.
    exists_mock = mocker.patch('os.path.exists', side_effect=[True, False, False])
    open_mock = mocker.patch('kopf._core.intents.piggybacking.open')
    open_mock.return_value.__enter__.return_value.read.side_effect=[' tkn ', RuntimeError]
    credentials = login_with_service_account()
    assert credentials is not None
    assert credentials.server == 'https://kubernetes.default.svc'
    assert credentials.default_namespace is None
    assert credentials.token == 'tkn'
    assert credentials.ca_path is None
    assert exists_mock.call_count == 3
    assert exists_mock.call_args_list[0][0][0] == SA_TOKEN_PATH
    assert exists_mock.call_args_list[1][0][0] == NAMESPACE_PATH
    assert exists_mock.call_args_list[2][0][0] == CA_PATH
    assert open_mock.call_count == 1
    assert open_mock.call_args_list[0][0][0] == SA_TOKEN_PATH



"""
The main Kopf module for all the exported functions & classes.
"""
# isort: skip_file

# Unlike all other places, where we import other modules and refer
# the functions via the modules, this is the framework's top-level interface,
# as it is seen by the users. So, we export the individual functions.

from kopf import (
    on,  # as a separate name on the public namespace
)
from kopf.on import (
    subhandler,
    register,
    daemon,
    timer,
    index,
)
from kopf._cogs.configs.configuration import (
    OperatorSettings,
)
from kopf._cogs.configs.diffbase import (
    DiffBaseStorage,
    AnnotationsDiffBaseStorage,
    StatusDiffBaseStorage,
    MultiDiffBaseStorage,
)
from kopf._cogs.configs.progress import (
    ProgressRecord,
    ProgressStorage,
    AnnotationsProgressStorage,
    StatusProgressStorage,
    MultiProgressStorage,
    SmartProgressStorage,
)
from kopf._cogs.helpers.typedefs import (
    Logger,
)
from kopf._cogs.helpers.versions import (
    version as __version__,
)
from kopf._cogs.structs.bodies import (
    RawEventType,
    RawEvent,
    RawBody,
    Status,
    Spec,
    Meta,
    Body,
    BodyEssence,
    Labels,
    Annotations,
    OwnerReference,
    ObjectReference,
    build_object_reference,
    build_owner_reference,
)
from kopf._cogs.structs.credentials import (
    LoginError,
    ConnectionInfo,
)
from kopf._cogs.structs.dicts import (
    FieldSpec,
    FieldPath,
)
from kopf._cogs.structs.diffs import (
    Diff,
    DiffItem,
    DiffOperation,
)
from kopf._cogs.structs.ephemera import (
    Memo,
    Index,
    Store,
)
from kopf._cogs.structs.ids import (
    HandlerId,
)
from kopf._cogs.structs.patches import (
    Patch,
)
from kopf._cogs.structs.references import (
    Resource,
    EVERYTHING,
)
from kopf._cogs.structs.reviews import (
    WebhookClientConfigService,
    WebhookClientConfig,
    Operation,
    UserInfo,
    Headers,
    SSLPeer,
    WebhookFn,
    WebhookServerProtocol,
)
from kopf._core.actions import (
    lifecycles,  # as a separate name on the public namespace
)
from kopf._core.actions.execution import (
    ErrorsMode,
    TemporaryError,
    PermanentError,
    HandlerTimeoutError,
    HandlerRetriesError,
)
from kopf._core.actions.lifecycles import (
    get_default_lifecycle,
    set_default_lifecycle,
)
from kopf._core.actions.loggers import (
    configure,
    LogFormat,
    ObjectLogger,
    LocalObjectLogger,
)
from kopf._core.engines.admission import (
    AdmissionError,
)
from kopf._core.engines.posting import (
    event,
    info,
    warn,
    exception,
)
from kopf._core.intents.callbacks import (
    not_,
    all_,
    any_,
    none_,
)
from kopf._core.intents.causes import (
    Reason,
)
from kopf._core.intents.filters import (
    ABSENT,
    PRESENT,
)
from kopf._core.intents.registries import (
    OperatorRegistry,
    get_default_registry,
    set_default_registry,
)
from kopf._core.intents.stoppers import (
    DaemonStopped,
    DaemonStoppingReason,
    SyncDaemonStopperChecker,  # deprecated
    AsyncDaemonStopperChecker,  # deprecated
)
from kopf._core.intents.piggybacking import (
    login_via_pykube,
    login_via_client,
    login_with_kubeconfig,
    login_with_service_account,
)
from kopf._core.reactor.running import (
    spawn_tasks,
    run_tasks,
    operator,
    run,
)
from kopf._core.reactor.subhandling import (
    execute,
)
from kopf._kits.hierarchies import (
    adopt,
    label,
    harmonize_naming,
    adjust_namespace,
    append_owner_reference,
    remove_owner_reference,
)
from kopf._kits.webhooks import (
    WebhookServer,
    WebhookK3dServer,
    WebhookMinikubeServer,
    WebhookNgrokTunnel,
    WebhookAutoServer,
    WebhookAutoTunnel,
)

__all__ = [
    'on', 'lifecycles', 'register', 'execute', 'daemon', 'timer', 'index',
    'configure', 'LogFormat',
    'login_via_pykube',
    'login_via_client',
    'login_with_kubeconfig',
    'login_with_service_account',
    'LoginError',
    'ConnectionInfo',
    'event', 'info', 'warn', 'exception',
    'spawn_tasks', 'run_tasks', 'operator', 'run',
    'adopt', 'label',
    'not_',
    'all_',
    'any_',
    'none_',
    'get_default_lifecycle', 'set_default_lifecycle',
    'build_object_reference', 'build_owner_reference',
    'append_owner_reference', 'remove_owner_reference',
    'ErrorsMode',
    'AdmissionError',
    'WebhookClientConfigService',
    'WebhookClientConfig',
    'Operation',
    'UserInfo',
    'Headers',
    'SSLPeer',
    'WebhookFn',
    'WebhookServerProtocol',
    'WebhookServer',
    'WebhookK3dServer',
    'WebhookMinikubeServer',
    'WebhookNgrokTunnel',
    'WebhookAutoServer',
    'WebhookAutoTunnel',
    'PermanentError',
    'TemporaryError',
    'HandlerTimeoutError',
    'HandlerRetriesError',
    'OperatorRegistry',
    'get_default_registry',
    'set_default_registry',
    'PRESENT', 'ABSENT',
    'OperatorSettings',
    'DiffBaseStorage',
    'AnnotationsDiffBaseStorage',
    'StatusDiffBaseStorage',
    'MultiDiffBaseStorage',
    'ProgressRecord',
    'ProgressStorage',
    'AnnotationsProgressStorage',
    'StatusProgressStorage',
    'MultiProgressStorage',
    'SmartProgressStorage',
    'RawEventType',
    'RawEvent',
    'RawBody',
    'Status',
    'Spec',
    'Meta',
    'Body',
    'BodyEssence',
    'Labels',
    'Annotations',
    'ObjectReference',
    'OwnerReference',
    'Memo', 'Index', 'Store',
    'Logger',
    'ObjectLogger',
    'LocalObjectLogger',
    'FieldSpec',
    'FieldPath',
    'Diff',
    'DiffItem',
    'DiffOperation',
    'HandlerId',
    'Reason',
    'Patch',
    'DaemonStopped',
    'DaemonStoppingReason',
    'SyncDaemonStopperChecker',  # deprecated
    'AsyncDaemonStopperChecker',  # deprecated
    'Resource', 'EVERYTHING',
]



import asyncio
from typing import Optional

import pytest

from kopf._cogs.aiokits import aiotasks, aiotoggles
from kopf._cogs.structs import bodies
from kopf._cogs.structs.references import Insights, Resource
from kopf._core.engines.peering import Identity
from kopf._core.reactor.orchestration import Ensemble, EnsembleKey, adjust_tasks


async def processor(*, raw_event: bodies.RawEvent, stream_pressure: Optional[asyncio.Event]) -> None:
    pass


@pytest.fixture(autouse=True)
def _auto_mocked(k8s_mocked):
    pass


@pytest.fixture()
async def insights(request, settings):
    insights = Insights()
    await insights.backbone.fill(resources=[
        request.getfixturevalue(name)
        for name in ['peering_resource', 'cluster_peering_resource', 'namespaced_peering_resource']
        if name in request.fixturenames
    ])
    settings.peering.mandatory = True
    return insights


@pytest.fixture()
async def ensemble(_no_asyncio_pending_tasks):
    operator_indexed = aiotoggles.ToggleSet(all)
    operator_paused = aiotoggles.ToggleSet(any)
    peering_missing = await operator_paused.make_toggle()
    ensemble = Ensemble(
        operator_indexed=operator_indexed,
        operator_paused=operator_paused,
        peering_missing=peering_missing,
    )

    try:
        yield ensemble
    finally:
        await aiotasks.stop(ensemble.get_tasks(ensemble.get_keys()), title='...')  # cleanup


async def test_empty_insights_cause_no_adjustments(
        settings, insights, ensemble):

    await adjust_tasks(
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert not ensemble.watcher_tasks
    assert not ensemble.peering_tasks
    assert not ensemble.pinging_tasks
    assert not ensemble.conflicts_found


async def test_new_resources_and_namespaces_spawn_new_tasks(
        settings, ensemble: Ensemble, insights: Insights, peering_resource):
    settings.peering.namespaced = peering_resource.namespaced

    r1 = Resource(group='group1', version='version1', plural='plural1', namespaced=True)
    r2 = Resource(group='group2', version='version2', plural='plural2', namespaced=True)
    insights.watched_resources.add(r1)
    insights.watched_resources.add(r2)
    insights.namespaces.add('ns1')
    insights.namespaces.add('ns2')
    r1ns1 = EnsembleKey(resource=r1, namespace='ns1')
    r1ns2 = EnsembleKey(resource=r1, namespace='ns2')
    r2ns1 = EnsembleKey(resource=r2, namespace='ns1')
    r2ns2 = EnsembleKey(resource=r2, namespace='ns2')
    peerns = peering_resource.namespaced
    peer1 = EnsembleKey(resource=peering_resource, namespace='ns1' if peerns else None)
    peer2 = EnsembleKey(resource=peering_resource, namespace='ns2' if peerns else None)

    await adjust_tasks(
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert set(ensemble.watcher_tasks) == {r1ns1, r1ns2, r2ns1, r2ns2}
    assert set(ensemble.peering_tasks) == {peer1, peer2}
    assert set(ensemble.pinging_tasks) == {peer1, peer2}
    assert set(ensemble.conflicts_found) == {peer1, peer2}


async def test_gone_resources_and_namespaces_stop_running_tasks(
        settings, ensemble: Ensemble, insights: Insights, peering_resource):
    settings.peering.namespaced = peering_resource.namespaced

    r1 = Resource(group='group1', version='version1', plural='plural1', namespaced=True)
    r2 = Resource(group='group2', version='version2', plural='plural2', namespaced=True)
    insights.watched_resources.add(r1)
    insights.watched_resources.add(r2)
    insights.namespaces.add('ns1')
    insights.namespaces.add('ns2')
    r1ns1 = EnsembleKey(resource=r1, namespace='ns1')
    r1ns2 = EnsembleKey(resource=r1, namespace='ns2')
    r2ns1 = EnsembleKey(resource=r2, namespace='ns1')
    r2ns2 = EnsembleKey(resource=r2, namespace='ns2')
    peerns = peering_resource.namespaced
    peer1 = EnsembleKey(resource=peering_resource, namespace='ns1' if peerns else None)

    await adjust_tasks(  # initialisation
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble
    )

    r1ns2_task = ensemble.watcher_tasks[r1ns2]
    r2ns1_task = ensemble.watcher_tasks[r2ns1]
    r2ns2_task = ensemble.watcher_tasks[r2ns2]

    insights.watched_resources.discard(r2)
    insights.namespaces.discard('ns2')

    await adjust_tasks(  # action-under-test
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert set(ensemble.watcher_tasks) == {r1ns1}
    assert set(ensemble.peering_tasks) == {peer1}
    assert set(ensemble.pinging_tasks) == {peer1}
    assert set(ensemble.conflicts_found) == {peer1}
    assert r1ns2_task.cancelled()
    assert r2ns1_task.cancelled()
    assert r2ns2_task.cancelled()


async def test_cluster_tasks_continue_running_on_namespace_deletion(
        settings, ensemble: Ensemble, insights: Insights, cluster_peering_resource):
    settings.peering.namespaced = cluster_peering_resource.namespaced

    r1 = Resource(group='group1', version='version1', plural='plural1', namespaced=True)
    r2 = Resource(group='group2', version='version2', plural='plural2', namespaced=True)
    insights.watched_resources.add(r1)
    insights.watched_resources.add(r2)
    insights.namespaces.add(None)
    r1nsN = EnsembleKey(resource=r1, namespace=None)
    r2nsN = EnsembleKey(resource=r2, namespace=None)
    peerN = EnsembleKey(resource=cluster_peering_resource, namespace=None)

    await adjust_tasks(  # initialisation
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble
    )

    r1nsN_task = ensemble.watcher_tasks[r1nsN]
    r2nsN_task = ensemble.watcher_tasks[r2nsN]

    insights.namespaces.discard(None)

    await adjust_tasks(  # action-under-test
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert set(ensemble.watcher_tasks) == {r1nsN, r2nsN}
    assert set(ensemble.peering_tasks) == {peerN}
    assert set(ensemble.pinging_tasks) == {peerN}
    assert set(ensemble.conflicts_found) == {peerN}
    assert not r1nsN_task.cancelled()
    assert not r2nsN_task.cancelled()
    assert not r1nsN_task.done()
    assert not r2nsN_task.done()


async def test_no_peering_tasks_with_no_peering_resources(
        settings, ensemble):

    settings.peering.mandatory = False
    insights = Insights()
    r1 = Resource(group='group1', version='version1', plural='plural1', namespaced=True)
    insights.watched_resources.add(r1)
    insights.namespaces.add('ns1')

    await adjust_tasks(
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert ensemble.watcher_tasks
    assert not ensemble.peering_tasks
    assert not ensemble.pinging_tasks
    assert not ensemble.conflicts_found


async def test_paused_with_mandatory_peering_but_absent_peering_resource(
        settings, ensemble: Ensemble):

    settings.peering.mandatory = True
    insights = Insights()

    await ensemble.peering_missing.turn_to(False)  # prerequisite
    assert ensemble.peering_missing.is_off()  # prerequisite
    assert ensemble.operator_paused.is_off()  # prerequisite

    await adjust_tasks(
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert ensemble.peering_missing.is_on()
    assert ensemble.operator_paused.is_on()


async def test_unpaused_with_mandatory_peering_and_existing_peering_resource(
        settings, ensemble: Ensemble, insights: Insights, peering_resource):
    settings.peering.namespaced = peering_resource.namespaced

    await ensemble.peering_missing.turn_to(True)  # prerequisite
    assert ensemble.peering_missing.is_on()  # prerequisite
    assert ensemble.operator_paused.is_on()  # prerequisite

    await adjust_tasks(
        processor=processor,
        identity=Identity('...'),
        settings=settings,
        insights=insights,
        ensemble=ensemble,
    )

    assert ensemble.peering_missing.is_off()
    assert ensemble.operator_paused.is_off()



import asyncio
import contextlib
from typing import Generator, Optional


@contextlib.contextmanager
def proper_loop(suggested_loop: Optional[asyncio.AbstractEventLoop] = None) -> Generator[None, None, None]:
    """
    Ensure that we have the proper loop, either suggested or properly managed.

    A "properly managed" loop is the one we own and therefore close.
    If ``uvloop`` is installed, it is used.
    Otherwise, the event loop policy remains unaffected.

    This loop manager is usually used in CLI only, not deeper than that;
    i.e. not even in ``kopf.run()``, since uvloop is only auto-managed for CLI.
    """
    original_policy = asyncio.get_event_loop_policy()
    if suggested_loop is None:  # the pure CLI use, not a KopfRunner or other code
        try:
            import uvloop
        except ImportError:
            pass
        else:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    try:
        yield

    finally:
        try:
            import uvloop
        except ImportError:
            pass
        else:
            asyncio.set_event_loop_policy(original_policy)



import pytest


@pytest.fixture(autouse=True)
def _autouse_resp_mocker(resp_mocker, version_api):
    pass


@pytest.fixture(autouse=True)
def _prevent_retries_in_api_tests(settings):
    settings.networking.error_backoffs = []



"""
Detecting the framework's own version.

The codebase does not contain the version directly, as it would require
code changes on every release. Kopf's releases depend on tagging rather
than in-code version bumps (Kopf's authour believes that versions belong
to the versioning system, not to the codebase).

The version is determined only once at startup when the code is loaded.
"""
from typing import Optional

version: Optional[str] = None

try:
    import importlib.metadata
except ImportError:
    pass
else:
    try:
        name, *_ = __name__.split('.')  # usually "kopf", unless renamed/forked.
        version = importlib.metadata.version(name)
    except Exception:
        pass  # installed as an egg, from git, etc.



import json
from typing import Any, Dict

import pytest

from kopf._cogs.clients.auth import APIContext, authenticated


def test_package_version():
    import kopf
    assert hasattr(kopf, '__version__')
    assert kopf.__version__  # not empty, not null


@pytest.mark.parametrize('version, useragent', [
    ('1.2.3', 'kopf/1.2.3'),
    ('1.2rc', 'kopf/1.2rc'),
    (None, 'kopf/unknown'),
])
async def test_http_user_agent_version(
        aresponses, hostname, fake_vault, mocker, version, useragent):

    mocker.patch('kopf._cogs.helpers.versions.version', version)

    @authenticated
    async def get_it(url: str, *, context: APIContext) -> Dict[str, Any]:
        response = await context.session.get(url)
        return await response.json()

    async def responder(request):
        return aresponses.Response(
            content_type='application/json',
            text=json.dumps(dict(request.headers)))

    aresponses.add(hostname, '/', 'get', responder)
    returned_headers = await get_it(f"http://{hostname}/")
    assert returned_headers['User-Agent'] == useragent
    await fake_vault.close()  # to prevent ResourceWarnings for unclosed connectors



"""
Keeping track of the cluster setup: namespaces, resources (custom and builtin).

The outcome of observation are "insights" -- a description of the cluster setup,
including the "backbone" -- core resources to be used by the operator/framework.

The resource specifications can be partial or even fuzzy (e.g. by categories),
with zero, one, or more actual resources matching the specification (selector).
Similarly, the namespaces can be specified by patterns or negations of them.

The actual resources & namespaces in the cluster are permanently observed
and matched against the specifications. Only those that do match are served.

If there are no permissions to observe the CRDs or namespaces, the framework
attempts the best possible fallback scenario:

* For CRDs, only the initial scan is done; no runtime observation is performed.
* For namespaces, the namespace patterns are treated as exact namespace names.

A warning is logged unless ``settings.scanning.disabled`` is set to true
to declare this restricted mode as the desired mode of operation.
"""
import asyncio
import functools
import logging
from typing import Collection, FrozenSet, Iterable, List, Optional, Set

from kopf._cogs.aiokits import aiotoggles
from kopf._cogs.clients import errors, fetching, scanning
from kopf._cogs.configs import configuration
from kopf._cogs.structs import bodies, references
from kopf._core.intents import handlers, registries
from kopf._core.reactor import queueing

logger = logging.getLogger(__name__)


async def namespace_observer(
        *,
        clusterwide: bool,
        namespaces: Collection[references.NamespacePattern],
        insights: references.Insights,
        settings: configuration.OperatorSettings,
) -> None:
    exact_namespaces = references.select_specific_namespaces(namespaces)
    resource = await insights.backbone.wait_for(references.NAMESPACES)

    # Populate the namespaces atomically (instead of notifying on every item from the watch-stream).
    if not settings.scanning.disabled and not clusterwide:
        try:
            objs, _ = await fetching.list_objs(
                settings=settings,
                resource=resource,
                namespace=None,
                logger=logger,
            )
            async with insights.revised:
                revise_namespaces(raw_bodies=objs, insights=insights, namespaces=namespaces)
                insights.revised.notify_all()
        except errors.APIForbiddenError:
            logger.warning("Not enough permissions to list namespaces. "
                           "Falling back to a list of namespaces which are assumed to exist: "
                           f"{exact_namespaces!r}")
            async with insights.revised:
                insights.namespaces.update(exact_namespaces)
                insights.revised.notify_all()
    else:
        async with insights.revised:
            insights.namespaces.update({None} if clusterwide else exact_namespaces)
            insights.revised.notify_all()

    # Notify those waiting for the initial listing (e.g. CLI commands).
    insights.ready_namespaces.set()

    if not settings.scanning.disabled and not clusterwide:
        try:
            await queueing.watcher(
                settings=settings,
                resource=resource,
                namespace=None,
                processor=functools.partial(process_discovered_namespace_event,
                                            namespaces=namespaces,
                                            insights=insights))
        except errors.APIForbiddenError:
            logger.warning("Not enough permissions to watch for namespaces: "
                           "changes (deletion/creation) will not be noticed; "
                           "the namespaces are only refreshed on operator restarts.")
            await asyncio.Event().wait()
    else:
        await asyncio.Event().wait()


async def resource_observer(
        *,
        settings: configuration.OperatorSettings,
        registry: registries.OperatorRegistry,
        insights: references.Insights,
) -> None:

    # Scan only the resource-related handlers, ignore activies & co.
    all_handlers: List[handlers.ResourceHandler] = []
    all_handlers.extend(registry._webhooks.get_all_handlers())
    all_handlers.extend(registry._indexing.get_all_handlers())
    all_handlers.extend(registry._watching.get_all_handlers())
    all_handlers.extend(registry._spawning.get_all_handlers())
    all_handlers.extend(registry._changing.get_all_handlers())
    groups = {handler.selector.group for handler in all_handlers if handler.selector is not None}
    groups.update({selector.group for selector in insights.backbone.selectors})

    # Prepopulate the resources before the dimension watchers start, so that each initially listed
    # namespace would start a watcher, and each initially listed CRD is already on the list.
    group_filter = None if None in groups else {group for group in groups if group is not None}
    resources = await scanning.scan_resources(groups=group_filter, settings=settings, logger=logger)
    async with insights.revised:
        revise_resources(resources=resources, insights=insights, registry=registry, group=None)
        await insights.backbone.fill(resources=resources)
        insights.revised.notify_all()

    # Notify those waiting for the initial listing (e.g. CLI commands).
    insights.ready_resources.set()

    # The resource watcher starts with an initial listing, and later reacts to any changes. However,
    # the existing resources are known already, so there will be no changes on the initial listing.
    resource = await insights.backbone.wait_for(references.CRDS)
    if not settings.scanning.disabled:
        try:
            await queueing.watcher(
                settings=settings,
                resource=resource,
                namespace=None,
                processor=functools.partial(process_discovered_resource_event,
                                            settings=settings,
                                            registry=registry,
                                            insights=insights))
        except errors.APIForbiddenError:
            logger.warning("Not enough permissions to watch for resources: "
                           "changes (creation/deletion/updates) will not be noticed; "
                           "the resources are only refreshed on operator restarts.")
            await asyncio.Event().wait()
    else:
        await asyncio.Event().wait()


async def process_discovered_namespace_event(
        *,
        raw_event: bodies.RawEvent,
        namespaces: Collection[references.NamespacePattern],
        insights: references.Insights,
        # Must be accepted whether used or not -- as passed by watcher()/worker().
        stream_pressure: Optional[asyncio.Event] = None,  # None for tests
        resource_indexed: Optional[aiotoggles.Toggle] = None,  # None for tests & observation
        operator_indexed: Optional[aiotoggles.ToggleSet] = None,  # None for tests & observation
) -> None:
    if raw_event['type'] is None:
        return

    async with insights.revised:
        revise_namespaces(raw_events=[raw_event], insights=insights, namespaces=namespaces)
        insights.revised.notify_all()


async def process_discovered_resource_event(
        *,
        raw_event: bodies.RawEvent,
        settings: configuration.OperatorSettings,
        registry: registries.OperatorRegistry,
        insights: references.Insights,
        # Must be accepted whether used or not -- as passed by watcher()/worker().
        stream_pressure: Optional[asyncio.Event] = None,  # None for tests
        resource_indexed: Optional[aiotoggles.Toggle] = None,  # None for tests & observation
        operator_indexed: Optional[aiotoggles.ToggleSet] = None,  # None for tests & observation
) -> None:
    # Ignore the initial listing, as all custom resources were already noticed by API listing.
    # This prevents numerous unneccessary API requests at the the start of the operator.
    if raw_event['type'] is None:
        return

    # Re-scan the whole dimension of resources if any single one of them changes. By this, we make
    # K8s's /apis/ endpoint the source of truth for all resources & versions & preferred versions,
    # instead of mimicking K8s in interpreting them ourselves (a probable source of bugs).
    # As long as it is atomic (for asyncio, i.e. sync), the existing tasks will not be affected.
    group = raw_event['object']['spec']['group']
    resources = await scanning.scan_resources(groups={group}, settings=settings, logger=logger)
    async with insights.revised:
        revise_resources(resources=resources, insights=insights, registry=registry, group=group)
        await insights.backbone.fill(resources=resources)
        insights.revised.notify_all()


def revise_namespaces(
        *,
        insights: references.Insights,
        namespaces: Collection[references.NamespacePattern],
        raw_events: Collection[bodies.RawEvent] = (),
        raw_bodies: Collection[bodies.RawBody] = (),
) -> None:
    all_events = list(raw_events) + [bodies.RawEvent(type=None, object=obj) for obj in raw_bodies]
    for raw_event in all_events:
        namespace = references.NamespaceName(raw_event['object']['metadata']['name'])
        matched = any(references.match_namespace(namespace, pattern) for pattern in namespaces)
        deleted = is_deleted(raw_event)
        if deleted:
            insights.namespaces.discard(namespace)
        elif matched:
            insights.namespaces.add(namespace)


def revise_resources(
        *,
        group: Optional[str],
        insights: references.Insights,
        registry: registries.OperatorRegistry,
        resources: Collection[references.Resource],
) -> None:

    # Scan only the resource-related handlers grouped by purpose; ignore activities & co.
    webhook_selectors = registry._webhooks.get_all_selectors()
    indexed_selectors = registry._indexing.get_all_selectors()
    watched_selectors = (
        registry._indexing.get_all_selectors() |
        registry._watching.get_all_selectors() |
        registry._spawning.get_all_selectors() |
        registry._changing.get_all_selectors()
    )
    patched_selectors = (
        registry._spawning.get_all_selectors() |
        registry._changing.get_all_selectors()
    )

    # Note: indexed and webhook resources are not checked for ambiguity or empty matching:
    # - the indexed resources are not served/watched directly and are only used as an utility cache;
    # - the webhook resources are PASSIVELY matched per HTTP request, so ambiguity is not a problem.
    # Ambiguity is a potential problem only for regular resource handlers because the operators
    # ACTIVELY trigger them and produce irreversible side-effects --- even if improperly configured.
    _update_resources(insights.webhook_resources, webhook_selectors, group=group, source=resources)
    _update_resources(insights.indexed_resources, indexed_selectors, group=group, source=resources)
    _update_resources(insights.watched_resources, watched_selectors, group=group, source=resources)
    _disable_ambiguous_selectors(resources=insights.watched_resources, selectors=watched_selectors)
    _disable_mismatched_selectors(resources=insights.watched_resources, selectors=watched_selectors)
    _disable_unsuitable_resources(resources=insights.watched_resources, selectors=patched_selectors)


def _update_resources(
        resources: Set[references.Resource],
        selectors: Iterable[references.Selector],
        *,
        group: Optional[str],
        source: Collection[references.Resource],
) -> None:
    """
    Update all or the group's resources from the source of resources.

    This also excludes the resources that continue to exist but stop matching
    the selectors: e.g. by category --- if a CRD's categories were modified.

    WARNING: We do not block the CRDs by adding finalizers (for simplicity),
    so it can be so that we miss the CRD deletion event and continue
    the watching attempts (and fail with HTTP 404).
    """

    # Exclude previously served resources that are gone now.
    group_resources = {resource for resource in resources if group in [None, resource.group]}
    resources.difference_update(group_resources)

    # Include or re-include the resources that are [still] served.
    for selector in selectors:
        resources.update(selector.select(source))


def _disable_ambiguous_selectors(
        *,
        resources: Set[references.Resource],
        selectors: Iterable[references.Selector],
) -> None:
    """
    Detect ambiguous selectors and stop serving/watching them.

    Ambiguous selectors are those matching 2+ distinct resources.
    For example, if a selector specifies "pods" and there are resources
    "pods.v1" and "pods.v1beta1.metrics.k8s.io" (but only if non-v1 resources
    cannot be filtered out completely; otherwise, implicitly prefer v1).
    """
    for selector in selectors:
        selected = selector.select(resources)
        if selector.is_specific and len(selected) > 1:
            logger.warning("Ambiguous resources will not be served (try specifying API groups):"
                           f" {selector} => {selected}")
            resources.difference_update(selected)


def _disable_mismatched_selectors(
        *,
        resources: Set[references.Resource],
        selectors: FrozenSet[references.Selector],
) -> None:
    """
    Warn for handlers that specify nonexistent resources.

    This can be due to a typo or a misconfiguration or CRDs are not yet created.
    """
    selector_names = ", ".join(
        f"{selector}"
        for selector in selectors
        if not selector.select(resources)
    )
    if selector_names:
        logger.warning("Unresolved resources cannot be served (try creating their CRDs):"
                       f" {selector_names}")


def _disable_unsuitable_resources(
        *,
        resources: Set[references.Resource],
        selectors: FrozenSet[references.Selector],
) -> None:

    # For both watching & patching, only look at watched resources, ignore webhook-only resources.
    nonwatchable_resources = {resource for resource in resources
                              if 'watch' not in resource.verbs or 'list' not in resource.verbs}
    nonpatchable_resources = {resource for resource in resources
                              if 'patch' not in resource.verbs} - nonwatchable_resources

    # For patching, only react if there are handlers that store a state (i.e. not on-event/index).
    patching_required = any(selector.select(nonpatchable_resources) for selector in selectors)

    if nonwatchable_resources:
        logger.warning(f"Non-watchable resources will not be served: {nonwatchable_resources}")
        resources.difference_update(nonwatchable_resources)
    if nonpatchable_resources and patching_required:
        logger.warning(f"Non-patchable resources will not be served: {nonpatchable_resources}")
        resources.difference_update(nonpatchable_resources)


def is_deleted(raw_event: bodies.RawEvent) -> bool:
    marked_as_deleted = bool(raw_event['object'].get('metadata', {}).get('deletionTimestamp'))
    really_is_deleted = raw_event['type'] == 'DELETED'
    return marked_as_deleted or really_is_deleted



"""
Send the custom events for the handled or other objects.
"""
import kopf


@kopf.on.create('kopfexamples')
def create_fn(body, **kwargs):

    # The all-purpose function for the event creation.
    kopf.event(body, type="SomeType", reason="SomeReason", message="Some message")

    # The shortcuts for the conventional events and common cases.
    kopf.info(body, reason="SomeReason", message="Some message")
    kopf.warn(body, reason="SomeReason", message="Some message")

    try:
        raise RuntimeError("Exception text.")
    except Exception:
        kopf.exception(body, reason="SomeReason", message="Some exception:")



import logging

import pytest

import kopf
from kopf._cogs.structs.bodies import Body
from kopf._cogs.structs.ephemera import Memo
from kopf._cogs.structs.patches import Patch
from kopf._core.actions.progression import State
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import ChangingCause, Reason


@pytest.mark.parametrize('lifecycle', [
    kopf.lifecycles.all_at_once,
    kopf.lifecycles.one_by_one,
    kopf.lifecycles.randomized,
    kopf.lifecycles.shuffled,
    kopf.lifecycles.asap,
])
async def test_protocol_invocation(lifecycle, resource):
    """
    To be sure that all kwargs are accepted properly.
    Especially when the new kwargs are added or an invocation protocol changed.
    """
    # The values are irrelevant, they can be anything.
    state = State.from_scratch()
    cause = ChangingCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=OperatorIndexers().indices,
        resource=resource,
        patch=Patch(),
        memo=Memo(),
        body=Body({}),
        initial=False,
        reason=Reason.NOOP,
    )
    handlers = []
    selected = lifecycle(handlers, state=state, **cause.kwargs)
    assert isinstance(selected, (tuple, list))
    assert len(selected) == 0



import kopf


@kopf.on.event('kopfexamples')
def event_fn_with_error(**kwargs):
    raise Exception("Oops!")


@kopf.on.event('kopfexamples')
def normal_event_fn(event, **kwargs):
    print(f"Event received: {event!r}")


# Marks for the e2e tests (see tests/e2e/test_examples.py):
E2E_ALLOW_TRACEBACKS = True
E2E_SUCCESS_COUNTS = {'normal_event_fn': 2}



import collections.abc
import contextlib
from contextvars import ContextVar
from typing import AsyncIterator, Iterable, Optional, Set

from kopf._cogs.configs import configuration
from kopf._cogs.structs import ids
from kopf._core.actions import execution, invocation, lifecycles, progression
from kopf._core.intents import callbacks, causes, handlers as handlers_, registries

# The task-local context; propagated down the stack instead of multiple kwargs.
# Used in `@kopf.subhandler` and `kopf.execute()` to add/get the sub-handlers.
subregistry_var: ContextVar[registries.ChangingRegistry] = ContextVar('subregistry_var')
subexecuted_var: ContextVar[bool] = ContextVar('subexecuted_var')


@contextlib.asynccontextmanager
async def subhandling_context() -> AsyncIterator[None]:
    with invocation.context([
        (subregistry_var, registries.ChangingRegistry()),
        (subexecuted_var, False),
    ]):
        # Go for normal handler invocation.
        yield

        # If the sub-handlers are not called explicitly, run them implicitly
        # as if it was done inside of the handler (still under the try-finally clause).
        if not subexecuted_var.get():
            await execute()


async def execute(
        *,
        fns: Optional[Iterable[callbacks.ChangingFn]] = None,
        handlers: Optional[Iterable[handlers_.ChangingHandler]] = None,
        registry: Optional[registries.ChangingRegistry] = None,
        lifecycle: Optional[execution.LifeCycleFn] = None,
        cause: Optional[execution.Cause] = None,
) -> None:
    """
    Execute the handlers in an isolated lifecycle.

    This function is just a public wrapper for `execute` with multiple
    ways to specify the handlers: either as the raw functions, or as the
    pre-created handlers, or as a registry (as used in the object handling).

    If no explicit functions or handlers or registry are passed,
    the sub-handlers of the current handler are assumed, as accumulated
    in the per-handler registry with ``@kopf.subhandler``.

    If the call to this method for the sub-handlers is not done explicitly
    in the handler, it is done implicitly after the handler is exited.
    One way or another, it is executed for the sub-handlers.
    """

    # Restore the current context as set in the handler execution cycle.
    lifecycle = lifecycle if lifecycle is not None else execution.sublifecycle_var.get()
    lifecycle = lifecycle if lifecycle is not None else lifecycles.get_default_lifecycle()
    cause = cause if cause is not None else execution.cause_var.get()
    parent_handler: execution.Handler = execution.handler_var.get()
    parent_prefix = parent_handler.id if parent_handler is not None else None

    # Validate the inputs; the function signatures cannot put these kind of restrictions, so we do.
    if len([v for v in [fns, handlers, registry] if v is not None]) > 1:
        raise TypeError("Only one of the fns, handlers, registry can be passed. Got more.")

    elif fns is not None and isinstance(fns, collections.abc.Mapping):
        subregistry = registries.ChangingRegistry()
        for id, fn in fns.items():
            real_id = registries.generate_id(fn=fn, id=id, prefix=parent_prefix)
            handler = handlers_.ChangingHandler(
                fn=fn, id=real_id, param=None,
                errors=None, timeout=None, retries=None, backoff=None,
                selector=None, labels=None, annotations=None, when=None,
                initial=None, deleted=None, requires_finalizer=None,
                reason=None, field=None, value=None, old=None, new=None,
                field_needs_change=None,
            )
            subregistry.append(handler)

    elif fns is not None and isinstance(fns, collections.abc.Iterable):
        subregistry = registries.ChangingRegistry()
        for fn in fns:
            real_id = registries.generate_id(fn=fn, id=None, prefix=parent_prefix)
            handler = handlers_.ChangingHandler(
                fn=fn, id=real_id, param=None,
                errors=None, timeout=None, retries=None, backoff=None,
                selector=None, labels=None, annotations=None, when=None,
                initial=None, deleted=None, requires_finalizer=None,
                reason=None, field=None, value=None, old=None, new=None,
                field_needs_change=None,
            )
            subregistry.append(handler)

    elif fns is not None:
        raise ValueError(f"fns must be a mapping or an iterable, got {fns.__class__}.")

    elif handlers is not None:
        subregistry = registries.ChangingRegistry()
        for handler in handlers:
            subregistry.append(handler)

    # Use the registry as is; assume that the caller knows what they do.
    elif registry is not None:
        subregistry = registry

    # Prevent double implicit execution.
    elif subexecuted_var.get():
        return

    # If no explicit args were passed, use the accumulated handlers from `@kopf.subhandler`.
    else:
        subexecuted_var.set(True)
        subregistry = subregistry_var.get()

    # The sub-handlers are only for upper-level causes, not for lower-level events.
    if not isinstance(cause, causes.ChangingCause):
        raise RuntimeError("Sub-handlers of event-handlers are not supported and have "
                           "no practical use (there are no retries or state tracking).")

    # Execute the real handlers (all or a few or one of them, as per the lifecycle).
    settings: configuration.OperatorSettings = execution.subsettings_var.get()
    owned_handlers = subregistry.get_resource_handlers(resource=cause.resource)
    cause_handlers = subregistry.get_handlers(cause=cause)
    storage = settings.persistence.progress_storage
    state = progression.State.from_storage(body=cause.body, storage=storage, handlers=owned_handlers)
    state = state.with_purpose(cause.reason).with_handlers(cause_handlers)
    outcomes = await execution.execute_handlers_once(
        lifecycle=lifecycle,
        settings=settings,
        handlers=cause_handlers,
        cause=cause,
        state=state,
        extra_context=subhandling_context,
    )
    state = state.with_outcomes(outcomes)
    state.store(body=cause.body, patch=cause.patch, storage=storage)
    progression.deliver_results(outcomes=outcomes, patch=cause.patch)

    # Enrich all parents with references to sub-handlers of any level deep (sub-sub-handlers, etc).
    # There is at least one container, as this function can be called only from a handler.
    subrefs_containers: Iterable[Set[ids.HandlerId]] = execution.subrefs_var.get()
    for key in state:
        for subrefs_container in subrefs_containers:
            subrefs_container.add(key)

    # Escalate `HandlerChildrenRetry` if the execute should be continued on the next iteration.
    if not state.done:
        raise execution.HandlerChildrenRetry(delay=state.delay)



import asyncio
import logging
from unittest.mock import Mock

import pytest

import kopf
from kopf._cogs.structs.ephemera import Memo
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import Reason
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event

EVENT_TYPES_WHEN_EXISTS = [None, 'ADDED', 'MODIFIED']


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_1st_level(registry, settings, resource, cause_mock, event_type,
                         caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.reason = Reason.CREATE

    fn_mock = Mock(return_value=None)
    sub1a_mock = Mock(return_value=None)
    sub1b_mock = Mock(return_value=None)

    # Only to justify the finalizer. See: cause_mock, which adds finalizers always.
    # TODO: get rid of mocks, test it normally.
    @kopf.on.delete('kopfexamples', id='del')
    async def _del(**_):
        pass

    @kopf.on.create('kopfexamples', id='fn')
    async def fn(**kwargs):
        fn_mock(**kwargs)
        @kopf.subhandler(id='sub1a')
        async def sub1a(**kwargs):
            sub1a_mock(**kwargs)
        @kopf.subhandler(id='sub1b')
        async def sub1b(**_):
            sub1b_mock(**kwargs)

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
    )

    assert fn_mock.call_count == 1
    assert sub1a_mock.call_count == 1
    assert sub1b_mock.call_count == 1

    assert k8s_mocked.sleep.call_count == 0
    assert k8s_mocked.patch.call_count == 1
    assert not event_queue.empty()

    assert_logs([
        "Creation is in progress:",
        "Handler 'fn' is invoked",
        "Handler 'fn/sub1a' is invoked",
        "Handler 'fn/sub1a' succeeded",
        "Handler 'fn/sub1b' is invoked",
        "Handler 'fn/sub1b' succeeded",
        "Handler 'fn' succeeded",
        "Creation is processed",
        "Patching with",
    ])


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_2nd_level(registry, settings, resource, cause_mock, event_type,
                         caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.reason = Reason.CREATE

    fn_mock = Mock(return_value=None)
    sub1a_mock = Mock(return_value=None)
    sub1b_mock = Mock(return_value=None)
    sub1a2a_mock = Mock(return_value=None)
    sub1a2b_mock = Mock(return_value=None)
    sub1b2a_mock = Mock(return_value=None)
    sub1b2b_mock = Mock(return_value=None)

    # Only to justify the finalizer. See: cause_mock, which adds finalizers always.
    # TODO: get rid of mocks, test it normally.
    @kopf.on.delete('kopfexamples', id='del')
    async def _del(**_):
        pass

    @kopf.on.create(*resource, id='fn')
    def fn(**kwargs):
        fn_mock(**kwargs)
        @kopf.subhandler(id='sub1a')
        def sub1a(**kwargs):
            sub1a_mock(**kwargs)
            @kopf.subhandler(id='sub1a2a')
            def sub1a2a(**kwargs):
                sub1a2a_mock(**kwargs)
            @kopf.subhandler(id='sub1a2b')
            def sub1a2b(**kwargs):
                sub1a2b_mock(**kwargs)
        @kopf.subhandler(id='sub1b')
        def sub1b(**kwargs):
            sub1b_mock(**kwargs)
            @kopf.subhandler(id='sub1b2a')
            def sub1b2a(**kwargs):
                sub1b2a_mock(**kwargs)
            @kopf.subhandler(id='sub1b2b')
            def sub1b2b(**kwargs):
                sub1b2b_mock(**kwargs)

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': {}},
        event_queue=event_queue,
    )

    assert fn_mock.call_count == 1
    assert sub1a_mock.call_count == 1
    assert sub1b_mock.call_count == 1
    assert sub1a2a_mock.call_count == 1
    assert sub1a2b_mock.call_count == 1
    assert sub1b2a_mock.call_count == 1
    assert sub1b2b_mock.call_count == 1

    assert k8s_mocked.sleep.call_count == 0
    assert k8s_mocked.patch.call_count == 1
    assert not event_queue.empty()

    assert_logs([
        "Creation is in progress:",
        "Handler 'fn' is invoked",
        "Handler 'fn/sub1a' is invoked",
        "Handler 'fn/sub1a/sub1a2a' is invoked",
        "Handler 'fn/sub1a/sub1a2a' succeeded",
        "Handler 'fn/sub1a/sub1a2b' is invoked",
        "Handler 'fn/sub1a/sub1a2b' succeeded",
        "Handler 'fn/sub1a' succeeded",
        "Handler 'fn/sub1b' is invoked",
        "Handler 'fn/sub1b/sub1b2a' is invoked",
        "Handler 'fn/sub1b/sub1b2a' succeeded",
        "Handler 'fn/sub1b/sub1b2b' is invoked",
        "Handler 'fn/sub1b/sub1b2b' succeeded",
        "Handler 'fn/sub1b' succeeded",
        "Handler 'fn' succeeded",
        "Creation is processed",
        "Patching with",
    ])



import pytest

import kopf
from kopf._core.intents.causes import HANDLER_REASONS, Reason


@pytest.mark.parametrize('deleted', [True, False, None])
@pytest.mark.parametrize('reason', HANDLER_REASONS)
def test_resumes_ignored_for_non_initial_causes(
        reason, deleted, cause_factory, resource):

    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=reason, initial=False,
                          body={'metadata': {'deletionTimestamp': '...'} if deleted else {}})

    @kopf.on.resume(*resource)
    def fn(**_):
        pass

    handlers = registry._changing.get_handlers(cause)
    assert len(handlers) == 0


@pytest.mark.parametrize('reason', list(set(HANDLER_REASONS) - {Reason.DELETE}))
def test_resumes_selected_for_initial_non_deletions(
        reason, cause_factory, resource):

    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=reason, initial=True)

    @kopf.on.resume(*resource)
    def fn(**_):
        pass

    handlers = registry._changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn


@pytest.mark.parametrize('reason', [Reason.DELETE])
def test_resumes_ignored_for_initial_deletions_by_default(
        reason, cause_factory, resource):

    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=reason, initial=True,
                          body={'metadata': {'deletionTimestamp': '...'}})

    @kopf.on.resume(*resource)
    def fn(**_):
        pass

    handlers = registry._changing.get_handlers(cause)
    assert len(handlers) == 0


@pytest.mark.parametrize('reason', [Reason.DELETE])
def test_resumes_selected_for_initial_deletions_when_explicitly_marked(
        reason, cause_factory, resource):

    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=reason, initial=True,
                          body={'metadata': {'deletionTimestamp': '...'}})

    @kopf.on.resume(*resource, deleted=True)
    def fn(**_):
        pass

    handlers = registry._changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn



import aiohttp.web
import pytest

from kopf._cogs.clients.errors import APIError
from kopf._cogs.clients.fetching import list_objs
from kopf._cogs.structs.credentials import LoginError


async def test_listing_works(
        resp_mocker, aresponses, hostname, settings, logger, resource, namespace,
        cluster_resource, namespaced_resource):

    result = {'items': [{}, {}]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    cluster_url = cluster_resource.get_url(namespace=None)
    namespaced_url = namespaced_resource.get_url(namespace='ns')
    aresponses.add(hostname, cluster_url, 'get', list_mock)
    aresponses.add(hostname, namespaced_url, 'get', list_mock)

    items, resource_version = await list_objs(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
    )
    assert items == result['items']

    assert list_mock.called
    assert list_mock.call_count == 1


# Note: 401 is wrapped into a LoginError and is tested elsewhere.
@pytest.mark.parametrize('status', [400, 403, 500, 666])
async def test_raises_direct_api_errors(
        resp_mocker, aresponses, hostname, settings, logger, status, resource, namespace,
        cluster_resource, namespaced_resource):

    list_mock = resp_mocker(return_value=aresponses.Response(status=status, reason='oops'))
    cluster_url = cluster_resource.get_url(namespace=None)
    namespaced_url = namespaced_resource.get_url(namespace='ns')
    aresponses.add(hostname, cluster_url, 'get', list_mock)
    aresponses.add(hostname, namespaced_url, 'get', list_mock)

    with pytest.raises(APIError) as e:
        await list_objs(
            logger=logger,
            settings=settings,
            resource=resource,
            namespace=namespace,
        )
    assert e.value.status == status



"""
Helpers for orchestrating asyncio tasks.

These utilities only support tasks, not more generic futures, coroutines,
or other awaitables. In most case where we use it, we need specifically tasks,
as we not only wait for them, but also cancel them.

Anyway, ``asyncio`` wraps all awaitables and coroutines into tasks on almost
all function calls with multiple awaiables (e.g. :func:`asyncio.wait`),
so there is no added overhead; instead, the implicit overhead is made explicit.
"""
import asyncio
from typing import TYPE_CHECKING, Any, Callable, Collection, Coroutine, \
                   NamedTuple, Optional, Set, Tuple, TypeVar

from kopf._cogs.helpers import typedefs

_T = TypeVar('_T')

# A workaround for a difference in tasks at runtime and type-checking time.
# Otherwise, at runtime: TypeError: 'type' object is not subscriptable.
if TYPE_CHECKING:
    Future = asyncio.Future[Any]
    Task = asyncio.Task[Any]
else:
    Future = asyncio.Future
    Task = asyncio.Task


async def cancel_coro(
        coro: Coroutine[Any, Any, Any],
        *,
        name: Optional[str] = None,
) -> None:
    """
    Cancel the coroutine if the wrapped code block is cancelled or fails.

    All coroutines must be awaited to prevent RuntimeWarnings/ResourceWarnings.
    As such, we generally need to create a dummy task and cancel it immediately.
    Despite asyncio tasks are lightweight, they still create an object and thus
    consume memory. This can be undesired when applied at scale --- e.g.
    in the multiplexer: when the watcher exits, it cancels all pending workers.

    To save memory, we first try to close the coroutine with no dummy task.
    As a fallback, the coroutine is cancelled gracefully via a dummy task.

    The context manager should be applied to all async code in the managing
    (i.e. parent/wrapping) coroutine from the beginning of it till the managed
    coro is actually awaited and executed.
    """
    try:
        # A dirty (undocumented) way to close a coro, but it saves memory.
        coro.close()  # OR: coro.throw(asyncio.CancelledError())
    except AttributeError:
        # The official way is to create an extra task object, thus to waste some memory.
        corotask = asyncio.create_task(coro=coro, name=name)
        corotask.cancel()
        try:
            await corotask
        except asyncio.CancelledError:
            pass  # cancellations are expected at this point


async def guard(
        coro: Coroutine[Any, Any, Any],
        name: str,
        *,
        flag: Optional[asyncio.Event] = None,
        finishable: bool = False,
        cancellable: bool = False,
        logger: Optional[typedefs.Logger] = None,
) -> None:
    """
    A guard for a presumably eternal (never-finishing) task.

    An "eternal" task is a task that never exits unless explicitly cancelled.
    If it does, this is a misbehaviour that is logged. Errors are always logged.
    Cancellations are also logged except if the task is said to be cancellable.

    It is used for background tasks that are started but never awaited/checked,
    so that the errors are not escalated properly; or if they are occasionally
    awaited/checked with a significant delay after an error possibly happend,
    but needs to be logged as soon as it happens.
    """
    capname = name.capitalize()

    # Guarded tasks can have prerequisites, which are set in other tasks.
    if flag is not None:
        try:
            await flag.wait()
        except asyncio.CancelledError:
            await cancel_coro(coro=coro, name=name)
            raise

    try:
        await coro
    except asyncio.CancelledError:
        if logger is not None and not cancellable:
            logger.debug(f"{capname} is cancelled.")
        raise
    except Exception as e:
        if logger is not None:
            logger.exception(f"{capname} has failed: {e}")
        raise
    else:
        if logger is not None and not finishable:
            logger.warning(f"{capname} has finished unexpectedly.")


def create_guarded_task(
        coro: Coroutine[Any, Any, Any],
        name: str,
        *,
        flag: Optional[asyncio.Event] = None,
        finishable: bool = False,
        cancellable: bool = False,
        logger: Optional[typedefs.Logger] = None,
) -> Task:
    """
    Create a guarded eternal task. See :func:`guard` for explanation.

    This is only a shortcut for named task creation (name is used in 2 places).
    """
    return asyncio.create_task(
        name=name,
        coro=guard(
            name=name,
            coro=coro,
            flag=flag,
            finishable=finishable,
            cancellable=cancellable,
            logger=logger))


async def wait(
        tasks: Collection[Task],
        *,
        timeout: Optional[float] = None,
        return_when: Any = asyncio.ALL_COMPLETED,
) -> Tuple[Set[Task], Set[Task]]:
    """
    A safer version of :func:`asyncio.wait` -- does not fail on an empty list.
    """
    if not tasks:
        return set(), set()
    done, pending = await asyncio.wait(tasks, timeout=timeout, return_when=return_when)
    return done, pending


async def stop(
        tasks: Collection[Task],
        *,
        title: str,
        quiet: bool = False,
        cancelled: bool = False,
        interval: Optional[float] = None,
        logger: Optional[typedefs.Logger] = None,
) -> Tuple[Set[Task], Set[Task]]:
    """
    Cancel the tasks and wait for them to finish; log if some are stuck.

    By default, all stopping activities are logged. In the quiet mode, logs
    only the tasks that are stuck longer than the designated polling interval,
    and their eventual exit if they were reported as stuck at least once.

    If the interval is not set, no polling is performed, and the stopping
    should happen in one iteration (even if it is going to take an eternity).

    The stopping itself does not have timeouts. It always ends either with
    the tasks stopped/exited, or with the stop-routine itself being cancelled.

    For better logging only, if the stopping routine is marked as performing
    the cancellation already (via ``cancelled=True``), then the cancellation
    of the stopping routine is considered as "double-cancelling".
    This does not affect the behaviour, but only the log messages.
    """
    captitle = title.capitalize()

    if not tasks:
        if logger is not None and not quiet:
            logger.debug(f"{captitle} tasks stopping is skipped: no tasks given.")
        return set(), set()

    for task in tasks:
        task.cancel()

    iterations = 0
    done_ever: Set[Task] = set()
    pending: Set[Task] = set(tasks)
    while pending:
        iterations += 1

        # If the waiting (current) task is cancelled before the wait is over,
        # propagate the cancellation to all the awaited (sub-) tasks, and let them finish.
        try:
            done_now, pending = await wait(pending, timeout=interval)
        except asyncio.CancelledError:
            # If the waiting (current) task is cancelled while propagating the cancellation
            # (i.e. double-cancelled), let it fail without graceful cleanup. It is urgent, it seems.
            pending = {task for task in tasks if not task.done()}
            if logger is not None and (not quiet or pending or iterations > 1):
                are = 'are' if not pending else 'are not'
                why = 'double-cancelling at stopping' if cancelled else 'cancelling at stopping'
                logger.debug(f"{captitle} tasks {are} stopped: {why}; tasks left: {pending!r}")
            raise  # the repeated cancellation, handled specially.
        else:
            # If the cancellation is propagated normally and the awaited (sub-) tasks exited,
            # consider it as a successful cleanup.
            if logger is not None and (not quiet or pending or iterations > 1):
                are = 'are' if not pending else 'are not'
                why = 'cancelling normally' if cancelled else 'finishing normally'
                logger.debug(f"{captitle} tasks {are} stopped: {why}; tasks left: {pending!r}")
            done_ever |= done_now

    return done_ever, pending


async def reraise(
        tasks: Collection[Task],
) -> None:
    """
    Re-raise errors from tasks, if any. Do nothing if all tasks have succeeded.
    """
    for task in tasks:
        try:
            task.result()  # can raise the regular (non-cancellation) exceptions.
        except asyncio.CancelledError:
            pass  # re-raise anything except regular cancellations/exits


async def all_tasks(
        *,
        ignored: Collection[Task] = frozenset(),
) -> Collection[Task]:
    """
    Return all tasks in the current event loop.

    Equivalent to :func:`asyncio.all_tasks`, but with an exlcusion list.
    The exclusion list is used to exclude the tasks that existed at a point
    in time in the past, to only get the tasks that appeared since then.
    """
    current_task = asyncio.current_task()
    return {task for task in asyncio.all_tasks()
            if task is not current_task and task not in ignored}


class SchedulerJob(NamedTuple):
    coro: Coroutine[Any, Any, Any]
    name: Optional[str]


class Scheduler:
    """
    An scheduler/orchestrator/executor for "fire-and-forget" tasks.

    Coroutines can be spawned via this scheduler and forgotten: no need to wait
    for them or to check their status --- the scheduler will take care of it.

    It is a simplified equivalent of aiojobs, but compatible with Python 3.10.
    Python 3.10 removed the explicit event loops (deprecated since Python 3.7),
    which broke aiojobs. At the same time, aiojobs looks unmaintained
    and contains no essential changes since July 2019 (i.e. for 2+ years).
    Hence the necessity to replicate the functionality.

    The scheduler is needed only for internal use and is not exposed to users.
    It is mainly used in the multiplexer (splitting a single stream of events
    of a resource kind into multiple queues of individual resource objects).
    Therefore, many of the features of aiojobs are removed as unnecessary:
    no individual task/job handling or closing, no timeouts, etc.

    .. note::

        Despite all coros will be wrapped into tasks sooner or later,
        and despite it is convincing to do this earlier and manage the tasks
        rather than our own queue of coros+names, do not do this:
        we want all tasks to refer to their true coros in their reprs,
        not to wrappers which wait until the running capacity is available.
    """

    def __init__(
            self,
            *,
            limit: Optional[int] = None,
            exception_handler: Optional[Callable[[BaseException], None]] = None,
    ) -> None:
        super().__init__()
        self._closed = False
        self._limit = limit
        self._exception_handler = exception_handler
        self._condition = asyncio.Condition()
        self._pending_coros: asyncio.Queue[SchedulerJob] = asyncio.Queue()
        self._running_tasks: Set[Task] = set()
        self._cleaning_queue: asyncio.Queue[Task] = asyncio.Queue()
        self._cleaning_task = asyncio.create_task(self._task_cleaner(), name=f"cleaner of {self!r}")
        self._spawning_task = asyncio.create_task(self._task_spawner(), name=f"spawner of {self!r}")

    def empty(self) -> bool:
        """ Check if the scheduler has nothing to do. """
        return self._pending_coros.empty() and not self._running_tasks

    async def wait(self) -> None:
        """
        Wait until the scheduler does nothing, i.e. idling (all tasks are done).
        """
        async with self._condition:
            await self._condition.wait_for(self.empty)

    async def close(self) -> None:
        """
        Stop accepting new tasks and cancel all running/pending ones.
        """

        # Running tasks are cancelled here. Pending tasks are cancelled at actual spawning.
        self._closed = True
        for task in self._running_tasks:
            task.cancel()

        # Wait until all tasks are fully done (it can take some time). This also includes
        # the pending coros, which are spawned and instantly cancelled (to prevent RuntimeWarnings).
        await self.wait()

        # Cleanup the scheduler's own resources.
        await stop({self._spawning_task, self._cleaning_task}, title="scheduler", quiet=True)

    async def spawn(
            self,
            coro: Coroutine[Any, Any, Any],
            *,
            name: Optional[str] = None,
    ) -> None:
        """
        Schedule a coroutine for ownership and eventual execution.

        Coroutine ownership ensures that all "fire-and-forget" coroutines
        that were passed to the scheduler will be awaited (to prevent warnings),
        even if the scheduler is closed before the coroutines are started.
        If a coroutine is added to a closed scheduler, it will be instantly
        cancelled before raising the scheduler's exception.
        """
        if self._closed:
            await cancel_coro(coro=coro, name=name)
            raise RuntimeError("Cannot add new coroutines to a closed and inactive scheduler.")
        async with self._condition:
            await self._pending_coros.put(SchedulerJob(coro=coro, name=name))
            self._condition.notify_all()  # -> task_spawner()

    def _can_spawn(self) -> bool:
        return (not self._pending_coros.empty() and
                (self._limit is None or len(self._running_tasks) < self._limit))

    async def _task_spawner(self) -> None:
        """ An internal meta-task to actually start pending coros as tasks. """
        while True:
            async with self._condition:
                await self._condition.wait_for(self._can_spawn)

                # Spawn as many tasks as allowed and as many coros as available at the moment.
                # Since nothing monitors the tasks "actively", we configure them to report back
                # when they are finished --- to be awaited and released "passively".
                while self._can_spawn():
                    coro, name = self._pending_coros.get_nowait()  # guaranteed by the predicate
                    task = asyncio.create_task(coro=coro, name=name)
                    task.add_done_callback(self._task_done_callback)
                    self._running_tasks.add(task)
                    if self._closed:
                        task.cancel()  # used to await the coros without executing them.

    async def _task_cleaner(self) -> None:
        """ An internal meta-task to cleanup the actually finished tasks. """
        while True:
            task = await self._cleaning_queue.get()

            # Await the task from an outer context to prevent RuntimeWarnings/ResourceWarnings.
            try:
                await task
            except BaseException:
                # The errors are handled in the done-callback. Suppress what has leaked for safety.
                pass

            # Ping other tasks to refill the pool of running tasks (or to close the scheduler).
            async with self._condition:
                self._running_tasks.discard(task)
                self._condition.notify_all()  # -> task_spawner() & close()

    def _task_done_callback(self, task: Task) -> None:
        # When a "fire-and-forget" task is done, release its system resources immediately:
        # nothing else is going to explicitly "await" for it any time soon, so we must do it.
        # But since a callback cannot be async, "awaiting" is done in a background utility task.
        self._running_tasks.discard(task)
        self._cleaning_queue.put_nowait(task)

        # If failed, initiate a callback defined by the owner of the task (if any).
        exc: Optional[BaseException]
        try:
            exc = task.exception()
        except asyncio.CancelledError:
            exc = None
        if exc is not None and self._exception_handler is not None:
            self._exception_handler(exc)



from typing import Type

import pytest

from kopf._cogs.configs.diffbase import AnnotationsDiffBaseStorage, \
                                        DiffBaseStorage, StatusDiffBaseStorage
from kopf._cogs.structs.bodies import Body

ALL_STORAGES = [AnnotationsDiffBaseStorage, StatusDiffBaseStorage]


@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_resource_references(
        cls: Type[DiffBaseStorage],
):
    body = Body({'apiVersion': 'group/version', 'kind': 'Kind'})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {}


@pytest.mark.parametrize('field', [
    'uid',
    'name',
    'namespace',
    'selfLink',
    'generation',
    'finalizers',
    'resourceVersion',
    'creationTimestamp',
    'deletionTimestamp',
    'any-unexpected-field',
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_system_fields_and_cleans_parents(
        field: str,
        cls: Type[DiffBaseStorage],
):
    body = Body({'metadata': {field: 'x'}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {}


@pytest.mark.parametrize('field', [
    'uid',
    'name',
    'namespace',
    'selfLink',
    'generation',
    'finalizers',
    'resourceVersion',
    'creationTimestamp',
    'deletionTimestamp',
    'any-unexpected-field',
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_system_fields_but_keeps_extra_fields(
        field: str,
        cls: Type[DiffBaseStorage],
):
    body = Body({'metadata': {field: 'x', 'other': 'y'}})
    storage = cls()
    essence = storage.build(body=body, extra_fields=['metadata.other'])
    assert essence == {'metadata': {'other': 'y'}}


@pytest.mark.parametrize('annotation', [
    pytest.param('kubectl.kubernetes.io/last-applied-configuration', id='kubectl'),
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_garbage_annotations_and_cleans_parents(
        annotation: str,
        cls: Type[DiffBaseStorage],
):
    body = Body({'metadata': {'annotations': {annotation: 'x'}}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {}


@pytest.mark.parametrize('annotation', [
    pytest.param('kubectl.kubernetes.io/last-applied-configuration', id='kubectl'),
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_garbage_annotations_but_keeps_others(
        annotation: str,
        cls: Type[DiffBaseStorage],
):
    body = Body({'metadata': {'annotations': {annotation: 'x', 'other': 'y'}}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {'metadata': {'annotations': {'other': 'y'}}}


@pytest.mark.parametrize('prefix', [
    'kopf-domain.tld',
    'kopf.domain.tld',
    'domain.tld.kopf',
    'domain.tld-kopf',
    'domain.kopf.tld',
    'domain.kopf-tld',
    'domain-kopf.tld',
    'domain-kopf-tld',
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_keeps_annotations_mentioning_kopf_but_not_from_other_operators(
        prefix: str,
        cls: Type[DiffBaseStorage],
):
    annotation = f'{prefix}/to-be-removed'
    body = Body({'metadata': {'annotations': {annotation: 'x'}}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {'metadata': {'annotations': {annotation: 'x'}}}


@pytest.mark.parametrize('prefix', [
    'kopf.zalando.org',
    'sub.kopf.zalando.org',
    'sub.sub.kopf.zalando.org',
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_other_operators_annotations_by_domain(
        prefix: str,
        cls: Type[DiffBaseStorage],
):
    annotation = f'{prefix}/to-be-removed'
    body = Body({'metadata': {'annotations': {annotation: 'x', 'other': 'y'}}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {'metadata': {'annotations': {'other': 'y'}}}


@pytest.mark.parametrize('prefix', [
    'domain.tld',
])
@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_other_operators_annotations_by_marker(
        prefix: str,
        cls: Type[DiffBaseStorage],
):
    marker = f'{prefix}/kopf-managed'
    annotation = f'{prefix}/to-be-removed'
    body = Body({'metadata': {'annotations': {annotation: 'x', marker: 'yes', 'other': 'y'}}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {'metadata': {'annotations': {'other': 'y'}}}


@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_status_and_cleans_parents(
        cls: Type[DiffBaseStorage],
):
    body = Body({'status': {'kopf': {'progress': 'x', 'anything': 'y'}, 'other': 'z'}})
    storage = cls()
    essence = storage.build(body=body)
    assert essence == {}


@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_removes_status_but_keeps_extra_fields(
        cls: Type[DiffBaseStorage],
):
    body = Body({'status': {'kopf': {'progress': 'x', 'anything': 'y'}, 'other': 'z'}})
    storage = cls()
    essence = storage.build(body=body, extra_fields=['status.other'])
    assert essence == {'status': {'other': 'z'}}


@pytest.mark.parametrize('cls', ALL_STORAGES)
def test_get_essence_clones_body(
        cls: Type[DiffBaseStorage],
):
    body = Body({'spec': {'depth': {'field': 'x'}}})
    storage = cls()
    essence = storage.build(body=body)
    body['spec']['depth']['field'] = 'y'
    assert essence is not body
    assert essence['spec'] is not body['spec']
    assert essence['spec']['depth'] is not body['spec']['depth']
    assert essence['spec']['depth']['field'] == 'x'



"""
A little note on how these tests work:

Almost all asyncio objects are not thread-safe, as per the official doc.
This includes `asyncio.Queue`. This queue is used for k8s-event posting.

K8s-events are posted via ``kopf.event()`` and similar calls,
and also via ``logger.info()`` for per-object logging messages.

The calls originate from various threads:

* Main thread where the framework's event-loop runs.
* Thread-pool executors for sync handlers.
* Explicitly started threads for object monitoring
  (e.g. from ``@kopf.on.resume`` handlers).

In the main thread, there is an event-loop running, and it has an asyncio task
to get the k8s-event events from the queue and to post them to the K8s API.

In the non-thread-safe mode, putting an object via `queue.put_nowait()``
does **NOT** wake up the pending ``queue.get()`` in the `poster` task
until something happens on the event-loop (not necessary on the queue).

In the thread-safe mode, putting an an object via `queue.put()``
(which is a coroutine and must be executed in the loop)
wakes the pending ``queue.get()`` call immediately.

These tests ensure that the thread-safe calls are used for k8s-event posting
by artificially reproducing the described situation. The delayed no-op task
(awakener) is used to wake up the event-loop after some time if the k8s-event
posting is not thread-safe. Otherwise, it wakes up on ``queue.get()`` instantly.

If thread safety is not ensured, the operators get sporadic errors regarding
thread-unsafe calls, which are difficult to catch and reproduce.
"""

import asyncio
import contextvars
import functools
import threading
import time

import pytest

from kopf import event

OBJ1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}}


@pytest.fixture()
def awakener():
    handles = []

    def noop():
        pass

    def awaken_fn(delay, fn=noop):
        handle = asyncio.get_running_loop().call_later(delay, fn)
        handles.append(handle)

    try:
        yield awaken_fn
    finally:
        for handle in handles:
            handle.cancel()


@pytest.fixture()
def threader():
    threads = []

    def start_fn(delay, fn):
        def thread_fn():
            time.sleep(delay)
            fn()

        target = functools.partial(contextvars.copy_context().run, thread_fn)
        thread = threading.Thread(target=target)
        thread.start()
        threads.append(thread)

    try:
        yield start_fn
    finally:
        for thread in threads:
            thread.join()


async def test_nonthreadsafe_indeed_fails(timer, awakener, threader, event_queue, event_queue_loop):

    def thread_fn():
        event_queue.put_nowait(object())

    awakener(0.7)
    threader(0.3, thread_fn)

    with timer:
        await event_queue.get()

    assert 0.6 <= timer.seconds <= 0.8


async def test_threadsafe_indeed_works(timer, awakener, threader, event_queue, event_queue_loop):

    def thread_fn():
        asyncio.run_coroutine_threadsafe(event_queue.put(object()), loop=event_queue_loop)

    awakener(0.7)
    threader(0.3, thread_fn)

    with timer:
        await event_queue.get()

    assert 0.2 <= timer.seconds <= 0.4


async def test_queueing_is_threadsafe(timer, awakener, threader, event_queue, event_queue_loop,
                                      settings_via_contextvar):

    def thread_fn():
        event(OBJ1, type='type1', reason='reason1', message='message1')

    awakener(0.7)
    threader(0.3, thread_fn)

    with timer:
        await event_queue.get()

    assert 0.2 <= timer.seconds <= 0.4



import asyncio
import collections.abc
from typing import Collection, Optional, Union


async def sleep(
        delays: Union[None, float, Collection[Union[None, float]]],
        wakeup: Optional[asyncio.Event] = None,
) -> Optional[float]:
    """
    Measure the sleep time: either until the timeout, or until the event is set.

    Returns the number of seconds left to sleep, or ``None`` if the sleep was
    not interrupted and reached its specified delay (an equivalent of ``0``).
    In theory, the result can be ``0`` if the sleep was interrupted precisely
    the last moment before timing out; this is unlikely to happen though.
    """
    passed_delays = delays if isinstance(delays, collections.abc.Collection) else [delays]
    actual_delays = [delay for delay in passed_delays if delay is not None]
    minimal_delay = min(actual_delays) if actual_delays else 0

    # Do not go for the real low-level system sleep if there is no need to sleep.
    if minimal_delay <= 0:
        return None

    awakening_event = wakeup if wakeup is not None else asyncio.Event()
    loop = asyncio.get_running_loop()
    try:
        start_time = loop.time()
        await asyncio.wait_for(awakening_event.wait(), timeout=minimal_delay)
    except asyncio.TimeoutError:
        return None  # interruptable sleep is over: uninterrupted.
    else:
        end_time = loop.time()
        duration = end_time - start_time
        return max(0., minimal_delay - duration)



import pytest

from kopf._kits.webhooks import ClusterDetector, WebhookAutoServer, WebhookAutoTunnel


# Reproducing the realistic environment would be costly and difficult,
# so we mock all external(!) libraries to return the results as we expect them.
# This reduces the quality of the tests, but makes them simple.
@pytest.fixture(autouse=True)
def pathmock(mocker, fake_vault, enforced_context, aresponses, hostname):
    mocker.patch('ssl.get_server_certificate', return_value='')
    mocker.patch('certvalidator.ValidationContext')
    validator = mocker.patch('certvalidator.CertificateValidator')
    pathmock = validator.return_value.validate_tls.return_value
    pathmock.first.issuer.native = {}
    pathmock.first.subject.native = {}
    return pathmock


async def test_no_detection(hostname, aresponses):
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.2.3'})
    hostname = await ClusterDetector().guess_host()
    assert hostname is None


async def test_dependencies(hostname, aresponses, no_certvalidator):
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.2.3'})
    with pytest.raises(ImportError) as err:
        await ClusterDetector().guess_host()
    assert "pip install certvalidator" in str(err.value)


async def test_minikube_via_issuer_cn(pathmock):
    pathmock.first.issuer.native = {'common_name': 'minikubeCA'}
    hostname = await ClusterDetector().guess_host()
    assert hostname == 'host.minikube.internal'


async def test_minikube_via_subject_cn(pathmock):
    pathmock.first.subject.native = {'common_name': 'minikube'}
    hostname = await ClusterDetector().guess_host()
    assert hostname == 'host.minikube.internal'


async def test_k3d_via_issuer_cn(pathmock):
    pathmock.first.issuer.native = {'common_name': 'k3s-ca-server-12345'}
    hostname = await ClusterDetector().guess_host()
    assert hostname == 'host.k3d.internal'


async def test_k3d_via_subject_cn(pathmock):
    pathmock.first.subject.native = {'common_name': 'k3s'}
    hostname = await ClusterDetector().guess_host()
    assert hostname == 'host.k3d.internal'


async def test_k3d_via_subject_org(pathmock):
    pathmock.first.subject.native = {'organization_name': 'k3s'}
    hostname = await ClusterDetector().guess_host()
    assert hostname == 'host.k3d.internal'


async def test_k3d_via_version_infix(hostname, aresponses):
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.20.4+k3s1'})
    hostname = await ClusterDetector().guess_host()
    assert hostname == 'host.k3d.internal'


async def test_server_detects(responder, aresponses, hostname, caplog, assert_logs):
    caplog.set_level(0)
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.20.4+k3s1'})
    server = WebhookAutoServer(insecure=True)
    async with server:
        async for _ in server(responder.fn):
            break  # do not sleep
    assert_logs(["Cluster detection found the hostname: host.k3d.internal"])


async def test_server_works(
        responder, aresponses, hostname, caplog, assert_logs):
    caplog.set_level(0)
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.20.4'})
    server = WebhookAutoServer(insecure=True)
    async with server:
        async for _ in server(responder.fn):
            break  # do not sleep
    assert_logs(["Cluster detection failed, running a simple local server"])


async def test_tunnel_detects(responder, pyngrok_mock, aresponses, hostname, caplog, assert_logs):
    caplog.set_level(0)
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.20.4+k3s1'})
    server = WebhookAutoTunnel()
    async with server:
        async for _ in server(responder.fn):
            break  # do not sleep
    assert_logs(["Cluster detection found the hostname: host.k3d.internal"])


async def test_tunnel_works(responder, pyngrok_mock, aresponses, hostname, caplog, assert_logs):
    caplog.set_level(0)
    aresponses.add(hostname, '/version', 'get', {'gitVersion': 'v1.20.4'})
    server = WebhookAutoTunnel()
    async with server:
        async for _ in server(responder.fn):
            break  # do not sleep
    assert_logs(["Cluster detection failed, using an ngrok tunnel."])



import logging

import kopf


async def test_daemon_is_spawned_at_least_once(
        resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle):
    caplog.set_level(logging.DEBUG)

    @kopf.daemon(*resource, id='fn')
    async def fn(**kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()

    await simulate_cycle({})

    await dummy.steps['called'].wait()
    await dummy.wait_for_daemon_done()

    assert dummy.mock.call_count == 1  # not restarted


async def test_daemon_initial_delay_obeyed(
        resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle):
    caplog.set_level(logging.DEBUG)

    @kopf.daemon(*resource, id='fn', initial_delay=1.0)
    async def fn(**kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()

    await simulate_cycle({})

    await dummy.steps['called'].wait()
    await dummy.wait_for_daemon_done()

    assert k8s_mocked.sleep.call_count >= 1
    assert k8s_mocked.sleep.call_count <= 2  # one optional extra call for sleep(None)
    assert k8s_mocked.sleep.call_args_list[0][0][0] == 1.0  # [call#][args/kwargs][arg#]



from kopf._cogs.structs.dicts import walk


def test_over_a_none():
    result = list(walk(None))
    assert len(result) == 0


def test_over_a_dict():
    obj = {}
    result = list(walk(obj))
    assert len(result) == 1
    assert result[0] is obj


def test_over_a_list_of_dicts():
    obj1 = {}
    obj2 = {}
    result = list(walk([obj1, obj2]))
    assert len(result) == 2
    assert result[0] is obj1
    assert result[1] is obj2


def test_over_a_tuple_of_dicts():
    obj1 = {}
    obj2 = {}
    result = list(walk((obj1, obj2)))
    assert len(result) == 2
    assert result[0] is obj1
    assert result[1] is obj2


def test_none_is_ignored():
    obj1 = {}
    obj2 = {}
    result = list(walk([obj1, None, obj2]))
    assert len(result) == 2
    assert result[0] is obj1
    assert result[1] is obj2


def test_simple_nested():
    obj1 = {'field': {'subfield': 'val'}}
    obj2 = {'field': {}}
    result = list(walk([obj1, obj2], nested=['field.subfield']))
    assert len(result) == 3
    assert result[0] is obj1
    assert result[1] == 'val'
    assert result[2] is obj2


def test_double_nested():
    obj1 = {'field': {'subfield': 'val'}}
    obj2 = {'field': {}}
    result = list(walk([obj1, obj2], nested=['field.subfield', 'field']))
    assert len(result) == 5
    assert result[0] is obj1
    assert result[1] == 'val'
    assert result[2] == {'subfield': 'val'}
    assert result[3] is obj2
    assert result[4] == {}



import logging.handlers
from typing import Collection

import pytest

from kopf._core.actions.loggers import LogFormat, ObjectFormatter, ObjectJsonFormatter, \
                                       ObjectPrefixingJsonFormatter, ObjectPrefixingTextFormatter, \
                                       ObjectTextFormatter, configure


@pytest.fixture(autouse=True)
def _clear_own_handlers():
    logger = logging.getLogger()
    logger.handlers[:] = [
        handler for handler in logger.handlers
        if not isinstance(handler, logging.StreamHandler) or
           not isinstance(handler.formatter, ObjectFormatter)
    ]
    original_handlers = logger.handlers[:]
    yield
    logger.handlers[:] = original_handlers


def _get_own_handlers(logger: logging.Logger) -> Collection[logging.Handler]:
    return [
        handler for handler in logger.handlers
        if isinstance(handler, logging.StreamHandler) and
           isinstance(handler.formatter, ObjectFormatter)
    ]


def test_own_formatter_is_used():
    configure()
    logger = logging.getLogger()
    own_handlers = _get_own_handlers(logger)
    assert len(own_handlers) == 1


@pytest.mark.parametrize('log_format', [LogFormat.FULL, LogFormat.PLAIN, '%(message)s'])
def test_formatter_nonprefixed_text(log_format):
    configure(log_format=log_format, log_prefix=False)
    logger = logging.getLogger()
    own_handlers = _get_own_handlers(logger)
    assert len(own_handlers) == 1
    assert type(own_handlers[0].formatter) is ObjectTextFormatter


@pytest.mark.parametrize('log_format', [LogFormat.FULL, LogFormat.PLAIN, '%(message)s'])
def test_formatter_prefixed_text(log_format):
    configure(log_format=log_format, log_prefix=True)
    logger = logging.getLogger()
    own_handlers = _get_own_handlers(logger)
    assert len(own_handlers) == 1
    assert type(own_handlers[0].formatter) is ObjectPrefixingTextFormatter


@pytest.mark.parametrize('log_format', [LogFormat.JSON])
def test_formatter_nonprefixed_json(log_format):
    configure(log_format=log_format, log_prefix=False)
    logger = logging.getLogger()
    own_handlers = _get_own_handlers(logger)
    assert len(own_handlers) == 1
    assert type(own_handlers[0].formatter) is ObjectJsonFormatter


@pytest.mark.parametrize('log_format', [LogFormat.JSON])
def test_formatter_prefixed_json(log_format):
    configure(log_format=log_format, log_prefix=True)
    logger = logging.getLogger()
    own_handlers = _get_own_handlers(logger)
    assert len(own_handlers) == 1
    assert type(own_handlers[0].formatter) is ObjectPrefixingJsonFormatter


@pytest.mark.parametrize('log_format', [LogFormat.JSON])
def test_json_has_no_prefix_by_default(log_format):
    configure(log_format=log_format, log_prefix=None)
    logger = logging.getLogger()
    own_handlers = _get_own_handlers(logger)
    assert len(own_handlers) == 1
    assert type(own_handlers[0].formatter) is ObjectJsonFormatter


def test_error_on_unknown_formatter():
    with pytest.raises(ValueError):
        configure(log_format=object())


@pytest.mark.parametrize('verbose, debug, quiet, expected_level', [
    (None, None, None, logging.INFO),
    (True, None, None, logging.DEBUG),
    (None, True, None, logging.DEBUG),
    (True, True, True, logging.DEBUG),
    (None, None, True, logging.WARNING),
])
def test_levels(verbose, debug, quiet, expected_level):
    configure(verbose=verbose, debug=debug, quiet=quiet)
    logger = logging.getLogger()
    assert logger.level == expected_level



import dataclasses

import aiohttp.web
import pytest

from kopf._cogs.clients.errors import APIError
from kopf._cogs.clients.patching import patch_obj
from kopf._cogs.structs.patches import Patch


async def test_without_subresources(
        resp_mocker, aresponses, hostname, settings, resource, namespace, logger):

    patch_mock = resp_mocker(return_value=aiohttp.web.json_response({}))
    aresponses.add(hostname, resource.get_url(namespace=namespace, name='name1'), 'patch', patch_mock)

    patch = Patch({'x': 'y'})
    await patch_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        patch=patch,
    )

    assert patch_mock.called
    assert patch_mock.call_count == 1

    data = patch_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'x': 'y'}


async def test_status_as_subresource_with_combined_payload(
        resp_mocker, aresponses, hostname, settings, resource, namespace, logger):
    resource = dataclasses.replace(resource, subresources=['status'])

    # Simulate Kopf's initial state and intention.
    patch = Patch({'spec': {'x': 'y'}, 'status': {'s': 't'}})

    # Simulate K8s API's behaviour. Assume something extra is added remotely.
    object_response = {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                       'spec': {'x': 'y', 'extra': '456'},
                       'status': '...'}
    status_response = {'status': {'s': 't', 'extra': '789'}}

    object_url = resource.get_url(namespace=namespace, name='name1')
    status_url = resource.get_url(namespace=namespace, name='name1', subresource='status')
    object_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(object_response))
    status_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(status_response))
    aresponses.add(hostname, object_url, 'patch', object_patch_mock)
    aresponses.add(hostname, status_url, 'patch', status_patch_mock)

    reconstructed = await patch_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        patch=patch,
    )

    assert object_patch_mock.called
    assert object_patch_mock.call_count == 1
    assert status_patch_mock.called
    assert status_patch_mock.call_count == 1

    data = object_patch_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'spec': {'x': 'y'}}
    data = status_patch_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'status': {'s': 't'}}

    assert reconstructed == {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                             'spec': {'x': 'y', 'extra': '456'},
                             'status': {'s': 't', 'extra': '789'}}


async def test_status_as_subresource_with_object_fields_only(
        resp_mocker, aresponses, hostname, settings, resource, namespace, logger):
    resource = dataclasses.replace(resource, subresources=['status'])

    # Simulate Kopf's initial state and intention.
    patch = Patch({'spec': {'x': 'y'}})

    # Simulate K8s API's behaviour. Assume something extra is added remotely.
    object_response = {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                       'spec': {'x': 'y', 'extra': '456'},
                       'status': '...'}
    status_response = {'status': {'s': 't', 'extra': '789'}}

    object_url = resource.get_url(namespace=namespace, name='name1')
    status_url = resource.get_url(namespace=namespace, name='name1', subresource='status')
    object_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(object_response))
    status_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(status_response))
    aresponses.add(hostname, object_url, 'patch', object_patch_mock)
    aresponses.add(hostname, status_url, 'patch', status_patch_mock)

    reconstructed = await patch_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        patch=patch,
    )

    assert object_patch_mock.called
    assert object_patch_mock.call_count == 1
    assert not status_patch_mock.called

    data = object_patch_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'spec': {'x': 'y'}}

    assert reconstructed == {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                             'spec': {'x': 'y', 'extra': '456'},
                             'status': '...'}


async def test_status_as_subresource_with_status_fields_only(
        resp_mocker, aresponses, hostname, settings, resource, namespace, logger):
    resource = dataclasses.replace(resource, subresources=['status'])

    # Simulate Kopf's initial state and intention.
    patch = Patch({'status': {'s': 't'}})

    # Simulate K8s API's behaviour. Assume something extra is added remotely.
    object_response = {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                       'spec': {'x': 'y', 'extra': '456'},
                       'status': '...'}
    status_response = {'status': {'s': 't', 'extra': '789'}}

    object_url = resource.get_url(namespace=namespace, name='name1')
    status_url = resource.get_url(namespace=namespace, name='name1', subresource='status')
    object_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(object_response))
    status_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(status_response))
    aresponses.add(hostname, object_url, 'patch', object_patch_mock)
    aresponses.add(hostname, status_url, 'patch', status_patch_mock)

    reconstructed = await patch_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        patch=patch,
    )

    assert not object_patch_mock.called
    assert status_patch_mock.called
    assert status_patch_mock.call_count == 1

    data = status_patch_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'status': {'s': 't'}}

    assert reconstructed == {'status': {'s': 't', 'extra': '789'}}


async def test_status_as_body_field_with_combined_payload(
        resp_mocker, aresponses, hostname, settings, resource, namespace, logger):

    # Simulate Kopf's initial state and intention.
    patch = Patch({'spec': {'x': 'y'}, 'status': {'s': 't'}})

    # Simulate K8s API's behaviour. Assume something extra is added remotely.
    object_response = {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                       'spec': {'x': 'y', 'extra': '456'},
                       'status': '...'}
    status_response = {'s': 't', 'extra': '789'}

    object_url = resource.get_url(namespace=namespace, name='name1')
    status_url = resource.get_url(namespace=namespace, name='name1', subresource='status')
    object_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(object_response))
    status_patch_mock = resp_mocker(return_value=aiohttp.web.json_response(status_response))
    aresponses.add(hostname, object_url, 'patch', object_patch_mock)
    aresponses.add(hostname, status_url, 'patch', status_patch_mock)

    reconstructed = await patch_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        patch=patch,
    )

    assert object_patch_mock.called
    assert object_patch_mock.call_count == 1
    assert not status_patch_mock.called

    data = object_patch_mock.call_args_list[0][0][0].data  # [callidx][args/kwargs][argidx]
    assert data == {'spec': {'x': 'y'}, 'status': {'s': 't'}}

    assert reconstructed == {'metadata': {'namespace': 'ns1', 'name': 'name1', 'extra': '123'},
                             'spec': {'x': 'y', 'extra': '456'},
                             'status': '...'}


@pytest.mark.parametrize('status', [404])
async def test_ignores_absent_objects(
        resp_mocker, aresponses, hostname, settings, status, resource, namespace, logger,
        cluster_resource, namespaced_resource):

    patch_mock = resp_mocker(return_value=aresponses.Response(status=status, reason='oops'))
    cluster_url = cluster_resource.get_url(namespace=None, name='name1')
    namespaced_url = namespaced_resource.get_url(namespace='ns', name='name1')
    aresponses.add(hostname, cluster_url, 'patch', patch_mock)
    aresponses.add(hostname, namespaced_url, 'patch', patch_mock)

    patch = {'x': 'y'}
    result = await patch_obj(
        logger=logger,
        settings=settings,
        resource=resource,
        namespace=namespace,
        name='name1',
        patch=patch,
    )

    assert result is None


# Note: 401 is wrapped into a LoginError and is tested elsewhere.
@pytest.mark.parametrize('status', [400, 403, 500, 666])
async def test_raises_api_errors(
        resp_mocker, aresponses, hostname, settings, status, resource, namespace, logger,
        cluster_resource, namespaced_resource):

    patch_mock = resp_mocker(return_value=aresponses.Response(status=status, reason='oops'))
    cluster_url = cluster_resource.get_url(namespace=None, name='name1')
    namespaced_url = namespaced_resource.get_url(namespace='ns', name='name1')
    aresponses.add(hostname, cluster_url, 'patch', patch_mock)
    aresponses.add(hostname, namespaced_url, 'patch', patch_mock)

    patch = {'x': 'y'}
    with pytest.raises(APIError) as e:
        await patch_obj(
            logger=logger,
            settings=settings,
            resource=resource,
            namespace=namespace,
            name='name1',
            patch=patch,
        )
    assert e.value.status == status



import functools
from typing import Any, AsyncGenerator, AsyncIterator, Callable, Dict, List, Tuple, TypeVar, cast

from kopf._cogs.structs import reviews

_SelfT = TypeVar('_SelfT')
_ServerFn = TypeVar('_ServerFn', bound=Callable[..., AsyncIterator[reviews.WebhookClientConfig]])


class WebhookContextManagerMeta(type):
    """
    Auto-decorate all ``__call__`` functions to persist their generators.

    Another way is via ``__init_subclass__``, but that requires monkey-patching
    and will not work on slotted classes (when they will be made slotted).
    """
    def __new__(
            cls,
            name: str,
            bases: Tuple[type, ...],
            namespace: Dict[str, Any],
            **kwargs: Any,
    ) -> "WebhookContextManagerMeta":
        if '__call__' in namespace:
            namespace['__call__'] = WebhookContextManager._persisted(namespace['__call__'])
        return super().__new__(cls, name, bases, namespace, **kwargs)


class WebhookContextManager(metaclass=WebhookContextManagerMeta):
    """
    A hack to make webhook servers/tunnels re-entrant.

    Generally, the webhook servers/tunnels are used only once per operator run,
    so the re-entrant servers/tunnels are not needed. This hack solves
    a problem that exists only in the unit-tests (because they are fast):

    When the garbage collection is postponed (as in PyPy), the server/tunnel
    frees the system resources (e.g. sockets) much later than the test is over
    (in ``finally:`` when the ``GeneratorExit`` is thrown into the generator).

    As a result, the resources (sockets) are released while the unrelated tests
    are running, and inject the exceptions into one of those unrelated tests
    (e.g. ``ResourceWarning: unclosed transport`` or alike).

    The obvious solution — the context managers — would break the protocol,
    which is promised to be a single callable that yields the client configs.

    To keep the backwards compatibility while cleaning up the resources on time:

    - Support the _optional_ async context manager protocol, _if_ implemented.
    - Implement a minimalistic context manager for the provided servers/tunnels.
    - Keep them fully functional when the context manager is not used.

    So, the servers/tunnels are left with the ``finally:`` block for cleanup.
    But the iterator-generator is remembered and force-closed on exit from
    the context manager — the same way as the garbage collector would close it.

    See more at :doc:`/admission`.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.__generators: List[AsyncGenerator[reviews.WebhookClientConfig, None]] = []

    async def __aenter__(self: _SelfT) -> _SelfT:
        return self

    async def __aexit__(self, *_: Any) -> None:
        # The order of cleanups is probably irrelevant, but prefer LIFO just in case:
        # the innermost and latest super() calls freed first, the outermost and earliest — last.
        for generator in reversed(self.__generators):
            try:
                await generator.aclose()
            except (GeneratorExit, StopAsyncIteration, StopIteration):
                pass
        self.__generators[:] = []

    @staticmethod
    def _persisted(wrapped: _ServerFn) -> _ServerFn:
        @functools.wraps(wrapped)
        async def wrapper(
                self: "WebhookContextManager",
                fn: reviews.WebhookFn,
        ) -> AsyncIterator[reviews.WebhookClientConfig]:
            iterator = wrapped(self, fn)
            if isinstance(iterator, AsyncGenerator):
                self.__generators.append(iterator)
            async for value in iterator:
                yield value
            if isinstance(iterator, AsyncGenerator):
                self.__generators.remove(iterator)
        return cast(_ServerFn, wrapper)



import asyncio
from unittest.mock import Mock

import pytest

from kopf._cogs.aiokits.aiotasks import Scheduler

CODE_OVERHEAD = 0.01


async def f(mock, *args):
    try:
        mock('started')
        for arg in args:
            if isinstance(arg, asyncio.Event):
                arg.set()
            elif isinstance(arg, float):
                await asyncio.sleep(arg)
            elif callable(arg):
                arg()
    except asyncio.CancelledError:
        mock('cancelled')
    else:
        mock('finished')


async def test_empty_scheduler_lifecycle(timer):
    with timer:
        scheduler = Scheduler()
        assert scheduler.empty()
        await scheduler.wait()
        assert scheduler.empty()
        await scheduler.close()
        assert scheduler.empty()
    assert timer.seconds < CODE_OVERHEAD


async def test_task_spawning_and_graceful_finishing(timer):
    mock = Mock()
    flag1 = asyncio.Event()
    flag2 = asyncio.Event()
    scheduler = Scheduler()

    result = await scheduler.spawn(f(mock, flag1, 0.1, flag2))
    assert result is None

    with timer:
        await flag1.wait()
    assert timer.seconds < CODE_OVERHEAD
    assert mock.call_args[0][0] == 'started'

    with timer:
        await flag2.wait()
    assert timer.seconds > 0.1
    assert timer.seconds < 0.1 + CODE_OVERHEAD
    assert mock.call_args[0][0] == 'finished'

    await scheduler.close()


async def test_task_spawning_and_cancellation(timer):
    mock = Mock()
    flag1 = asyncio.Event()
    flag2 = asyncio.Event()
    scheduler = Scheduler()

    result = await scheduler.spawn(f(mock, flag1, 1.0, flag2))
    assert result is None

    with timer:
        await flag1.wait()
    assert timer.seconds < CODE_OVERHEAD
    assert mock.call_args[0][0] == 'started'

    with timer:
        await scheduler.close()
    assert timer.seconds < CODE_OVERHEAD  # near-instant
    assert mock.call_args[0][0] == 'cancelled'


async def test_no_tasks_are_accepted_after_closing():
    scheduler = Scheduler()
    await scheduler.close()

    assert scheduler._closed
    assert scheduler._spawning_task.done()
    assert scheduler._cleaning_task.done()

    with pytest.raises(RuntimeError, match=r"Cannot add new coroutines"):
        await scheduler.spawn(f(Mock(), 1.0))


async def test_successes_are_not_reported():
    exception_handler = Mock()
    scheduler = Scheduler(exception_handler=exception_handler)
    await scheduler.spawn(f(Mock()))
    await scheduler.wait()
    await scheduler.close()
    assert exception_handler.call_count == 0


async def test_cancellations_are_not_reported():
    exception_handler = Mock()
    mock = Mock(side_effect=asyncio.CancelledError())
    scheduler = Scheduler(exception_handler=exception_handler)
    await scheduler.spawn(f(mock, 1))
    await scheduler.wait()
    await scheduler.close()
    assert exception_handler.call_count == 0


async def test_exceptions_are_reported():
    exception = ValueError('hello')
    exception_handler = Mock()
    mock = Mock(side_effect=exception)
    scheduler = Scheduler(exception_handler=exception_handler)
    await scheduler.spawn(f(mock))
    await scheduler.wait()
    await scheduler.close()
    assert exception_handler.call_count == 1
    assert exception_handler.call_args[0][0] is exception


async def test_tasks_are_parallel_if_limit_is_not_reached(timer):
    """
    time:  ////////----------------------0.1s------------------0.2s--///
    task1: ->spawn->start->sleep->finish->|
    task2: ->spawn->start->sleep->finish->|
    """
    task1_started = asyncio.Event()
    task1_finished = asyncio.Event()
    task2_started = asyncio.Event()
    task2_finished = asyncio.Event()
    scheduler = Scheduler(limit=2)

    with timer:
        await scheduler.spawn(f(Mock(), task1_started, 0.1, task1_finished))
        await scheduler.spawn(f(Mock(), task2_started, 0.1, task2_finished))
    assert timer.seconds < CODE_OVERHEAD  # i.e. spawning is not not blocking

    with timer:
        await task1_finished.wait()
        assert task2_started.is_set()
        await task2_finished.wait()

    assert timer.seconds > 0.1
    assert timer.seconds < 0.1 + CODE_OVERHEAD

    await scheduler.close()


async def test_tasks_are_pending_if_limit_is_reached(timer):
    """
    time:  ////////----------------------0.1s------------------0.2s--///
    task1: ->spawn->start->sleep->finish->|
    task2: ->spawn->.....(pending)......->start->sleep->finish->|
    """
    task1_started = asyncio.Event()
    task1_finished = asyncio.Event()
    task2_started = asyncio.Event()
    task2_finished = asyncio.Event()
    scheduler = Scheduler(limit=1)

    with timer:
        await scheduler.spawn(f(Mock(), task1_started, 0.1, task1_finished))
        await scheduler.spawn(f(Mock(), task2_started, 0.1, task2_finished))
    assert timer.seconds < CODE_OVERHEAD  # i.e. spawning is not not blocking

    with timer:
        await task1_finished.wait()
        assert not task2_started.is_set()
        await task2_finished.wait()

    assert timer.seconds > 0.2
    assert timer.seconds < 0.2 + CODE_OVERHEAD * 2

    await scheduler.close()



import asyncio
import datetime
import random
from typing import Dict

import kopf
import pykube

LOCK: asyncio.Lock  # requires a loop on creation
STOPPERS: Dict[str, Dict[str, asyncio.Event]] = {}  # [namespace][name]


@kopf.on.startup()
async def startup_fn_simple(logger, **kwargs):
    logger.info("Initialising the task-lock...")
    global LOCK
    LOCK = asyncio.Lock()  # in the current asyncio loop


@kopf.on.startup()
async def startup_fn_retried(retry, logger, **kwargs):
    if retry < 3:
        raise kopf.TemporaryError(f"Going to succeed in {3-retry}s", delay=1)
    else:
        logger.info("Starting retried...")
        # raise kopf.PermanentError("Unable to start!")


@kopf.on.cleanup()
async def cleanup_fn(logger, **kwargs):
    logger.info("Cleaning up...")
    for namespace in STOPPERS.keys():
        for name, flag in STOPPERS[namespace].items():
            flag.set()
    logger.info("All pod-tasks are requested to stop...")


@kopf.on.login(errors=kopf.ErrorsMode.PERMANENT)
async def login_fn(**kwargs):
    print('Logging in in 2s...')
    await asyncio.sleep(2.0)

    # An equivalent of kopf.login_via_pykube(), but shrinked for demo purposes.
    config = pykube.KubeConfig.from_env()
    ca = config.cluster.get('certificate-authority')
    cert = config.user.get('client-certificate')
    pkey = config.user.get('client-key')
    return kopf.ConnectionInfo(
        server=config.cluster.get('server'),
        ca_path=ca.filename() if ca else None,  # can be a temporary file
        insecure=config.cluster.get('insecure-skip-tls-verify'),
        username=config.user.get('username'),
        password=config.user.get('password'),
        token=config.user.get('token'),
        certificate_path=cert.filename() if cert else None,  # can be a temporary file
        private_key_path=pkey.filename() if pkey else None,  # can be a temporary file
        default_namespace=config.namespace,
        expiration=datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(seconds=30),
    )


@kopf.on.probe()
async def tasks_count(**kwargs):
    return sum(len(flags) for flags in STOPPERS.values())


@kopf.on.probe()
async def monitored_objects(**kwargs):
    return {namespace: sorted(name for name in STOPPERS[namespace]) for namespace in STOPPERS}


@kopf.on.event('pods')
async def pod_task(namespace, name, logger, **_):
    async with LOCK:
        if namespace not in STOPPERS or name not in STOPPERS[namespace]:
            flag = asyncio.Event()
            STOPPERS.setdefault(namespace, {}).setdefault(name, flag)
            asyncio.create_task(_task_fn(logger, shouldstop=flag))


async def _task_fn(logger, shouldstop: asyncio.Event):
    while not shouldstop.is_set():
        await asyncio.sleep(random.randint(1, 10))
        logger.info("Served by the background task.")
    logger.info("Serving is finished by request.")


# Marks for the e2e tests (see tests/e2e/test_examples.py):
E2E_STARTUP_STOP_WORDS = ['Served by the background task.']
E2E_CLEANUP_STOP_WORDS = ['Hung tasks', 'Root tasks']
E2E_SUCCESS_COUNTS = {'startup_fn_simple': 1, 'startup_fn_retried': 1, 'cleanup_fn': 1}



"""
Admission reviews: requests & responses, also the webhook server protocols.
"""
from typing import Any, AsyncIterator, Awaitable, Callable, List, Mapping, Optional, Union

from typing_extensions import Literal, Protocol, TypedDict

from kopf._cogs.structs import bodies

Headers = Mapping[str, str]
SSLPeer = Mapping[str, Any]

Operation = Literal['CREATE', 'UPDATE', 'DELETE', 'CONNECT']


class RequestKind(TypedDict):
    group: str
    version: str
    kind: str


class RequestResource(TypedDict):
    group: str
    version: str
    resource: str


class UserInfo(TypedDict):
    username: str
    uid: str
    groups: List[str]


class CreateOptions(TypedDict, total=False):
    apiVersion: Literal["meta.k8s.io/v1"]
    kind: Literal["CreateOptions"]


class UpdateOptions(TypedDict, total=False):
    apiVersion: Literal["meta.k8s.io/v1"]
    kind: Literal["UpdateOptions"]


class DeleteOptions(TypedDict, total=False):
    apiVersion: Literal["meta.k8s.io/v1"]
    kind: Literal["DeleteOptions"]


class RequestPayload(TypedDict):
    uid: str
    kind: RequestKind
    resource: RequestResource
    subResource: Optional[str]
    requestKind: RequestKind
    requestResource: RequestResource
    requestSubResource: Optional[str]
    userInfo: UserInfo
    name: str
    namespace: Optional[str]
    operation: Operation
    options: Union[None, CreateOptions, UpdateOptions, DeleteOptions]
    dryRun: bool
    object: bodies.RawBody
    oldObject: Optional[bodies.RawBody]


class Request(TypedDict):
    apiVersion: Literal["admission.k8s.io/v1", "admission.k8s.io/v1beta1"]
    kind: Literal["AdmissionReview"]
    request: RequestPayload


class ResponseStatus(TypedDict, total=False):
    code: int
    message: str


class ResponsePayload(TypedDict, total=False):
    uid: str
    allowed: bool
    warnings: Optional[List[str]]
    status: Optional[ResponseStatus]
    patch: Optional[str]
    patchType: Optional[Literal["JSONPatch"]]


class Response(TypedDict):
    apiVersion: Literal["admission.k8s.io/v1", "admission.k8s.io/v1beta1"]
    kind: Literal["AdmissionReview"]
    response: ResponsePayload


class WebhookClientConfigService(TypedDict, total=False):
    namespace: Optional[str]
    name: Optional[str]
    path: Optional[str]
    port: Optional[int]


class WebhookClientConfig(TypedDict, total=False):
    """
    A config of clients (apiservers) to access the webhooks' server (operators).

    This dictionary is put into managed webhook configurations "as is".
    The fields & type annotations are only for hinting.

    Kopf additionally modifies the url and the service's path to inject
    handler ids as the last path component. This must be taken into account
    by custom webhook servers.
    """
    caBundle: Optional[str]  # if absent, the default apiservers' trust chain is used.
    url: Optional[str]
    service: Optional[WebhookClientConfigService]


class WebhookFn(Protocol):
    """
    A framework-provided function to call when an admission request is received.

    The framework provides the actual function. Custom webhook servers must
    accept the function, invoke it accordingly on admission requests, wait
    for the admission response, serialise it and send it back. They do not
    implement this function. This protocol only declares the exact signature.
    """
    def __call__(
            self,
            request: Request,
            *,
            webhook: Optional[str] = None,
            headers: Optional[Mapping[str, str]] = None,
            sslpeer: Optional[Mapping[str, Any]] = None,
    ) -> Awaitable[Response]: ...


# A server (either a coroutine or a callable object).
WebhookServerProtocol = Callable[[WebhookFn], AsyncIterator[WebhookClientConfig]]



import os.path
import subprocess
import time

import kopf.testing
import pytest

crd_yaml = os.path.relpath(os.path.join(os.path.dirname(__file__), '..', 'crd.yaml'))
obj_yaml = os.path.relpath(os.path.join(os.path.dirname(__file__), '..', 'obj.yaml'))
example_py = os.path.relpath(os.path.join(os.path.dirname(__file__), 'example.py'))


@pytest.fixture(autouse=True)
def crd_exists():
    subprocess.run(f"kubectl apply -f {crd_yaml}",
                   check=True, timeout=10, capture_output=True, shell=True)


@pytest.fixture(autouse=True)
def obj_absent():
    # Operator is not running in fixtures, so we need a force-delete (or this patch).
    subprocess.run(['kubectl', 'patch', '-f', obj_yaml,
                    '-p', '{"metadata":{"finalizers":[]}}',
                    '--type', 'merge'],
                   check=False, timeout=10, capture_output=True)
    subprocess.run(f"kubectl delete -f {obj_yaml}",
                   check=False, timeout=10, capture_output=True, shell=True)


def test_handler_filtering():

    # To prevent lengthy threads in the loop executor when the process exits.
    settings = kopf.OperatorSettings()
    settings.watching.server_timeout = 10

    # Run an operator and simulate some activity with the operated resource.
    with kopf.testing.KopfRunner(
        ['run', '--all-namespaces', '--verbose', '--standalone', example_py],
        settings=settings,
    ) as runner:

        subprocess.run(f"kubectl create -f {obj_yaml}",
                       shell=True, check=True, timeout=10, capture_output=True)
        time.sleep(5)  # give it some time to react
        subprocess.run(f"kubectl patch -f {obj_yaml} --type merge -p '" '{"spec":{"field":"changed"}}' "'",
                       shell=True, check=True, timeout=10, capture_output=True)
        time.sleep(2)  # give it some time to react
        subprocess.run(f"kubectl delete -f {obj_yaml}",
                       shell=True, check=True, timeout=10, capture_output=True)
        time.sleep(1)  # give it some time to react

    # Ensure that the operator did not die on start, or during the operation.
    assert runner.exception is None
    assert runner.exit_code == 0

    # Check for correct log lines (to indicate correct handlers were executed).
    assert '[default/kopf-example-1] Label is matching.' in runner.stdout
    assert '[default/kopf-example-1] Label is present.' in runner.stdout
    assert '[default/kopf-example-1] Label is absent.' in runner.stdout
    assert '[default/kopf-example-1] Label callback matching.' in runner.stdout
    assert '[default/kopf-example-1] Annotation is matching.' in runner.stdout
    assert '[default/kopf-example-1] Annotation is present.' in runner.stdout
    assert '[default/kopf-example-1] Annotation is absent.' in runner.stdout
    assert '[default/kopf-example-1] Annotation callback mismatch.' not in runner.stdout
    assert '[default/kopf-example-1] Filter satisfied.' in runner.stdout
    assert '[default/kopf-example-1] Filter not satisfied.' not in runner.stdout
    assert '[default/kopf-example-1] Field value is satisfied.' in runner.stdout
    assert '[default/kopf-example-1] Field value is not satisfied.' not in runner.stdout
    assert '[default/kopf-example-1] Field presence is satisfied.' in runner.stdout
    assert '[default/kopf-example-1] Field presence is not satisfied.' not in runner.stdout
    assert '[default/kopf-example-1] Field change is satisfied.' in runner.stdout
    assert '[default/kopf-example-1] Field daemon is satisfied.' in runner.stdout



"""
Execution of pre-selected handlers, in batches or individually.

These functions are invoked from `kopf._core.reactor.processing`,
where the raw watch-events are interpreted and wrapped into extended *causes*.

The handler execution can also be used in other places, such as in-memory
activities, when there is no underlying Kubernetes object to patch'n'watch.
"""
import asyncio
import contextlib
import dataclasses
import datetime
import enum
from contextvars import ContextVar
from typing import Any, AsyncContextManager, AsyncIterator, Callable, Collection, Iterable, \
                   Mapping, MutableMapping, NewType, Optional, Sequence, Set, TypeVar

from typing_extensions import Protocol

from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import ids
from kopf._core.actions import invocation

# The default delay duration for the regular exception in retry-mode.
DEFAULT_RETRY_DELAY = 1 * 60


class PermanentError(Exception):
    """ A fatal handler error, the retries are useless. """


class TemporaryError(Exception):
    """ A potentially recoverable error, should be retried. """
    def __init__(
            self,
            __msg: Optional[str] = None,
            delay: Optional[float] = DEFAULT_RETRY_DELAY,
    ) -> None:
        super().__init__(__msg)
        self.delay = delay


class HandlerTimeoutError(PermanentError):
    """ An error for the handler's timeout (if set). """


class HandlerRetriesError(PermanentError):
    """ An error for the handler's retries exceeded (if set). """


class HandlerChildrenRetry(TemporaryError):
    """ An internal pseudo-error to retry for the next sub-handlers attempt. """


class ErrorsMode(enum.Enum):
    """ How arbitrary (non-temporary/non-permanent) exceptions are treated. """
    IGNORED = enum.auto()
    TEMPORARY = enum.auto()
    PERMANENT = enum.auto()


# A specialised type to highlight the purpose or origin of the data of type Any,
# to not be mixed with other arbitrary Any values, where it is indeed "any".
Result = NewType('Result', object)


@dataclasses.dataclass(frozen=True)
class Outcome:
    """
    An in-memory outcome of one single invocation of one single handler.

    Conceptually, an outcome is similar to the async futures, but some cases
    are handled specially: e.g., the temporary errors have exceptions,
    but the handler should be retried later, unlike with the permanent errors.

    Note the difference: `HandlerState` is a persistent state of the handler,
    possibly after a few executions, and consisting of simple data types
    (for YAML/JSON serialisation) rather than the actual in-memory objects.
    """
    final: bool
    delay: Optional[float] = None
    result: Optional[Result] = None
    exception: Optional[Exception] = None
    subrefs: Collection[ids.HandlerId] = ()


@dataclasses.dataclass(frozen=True)
class HandlerState:
    """
    A persisted state of a single handler, as stored on the resource's status.

    Note the difference: `Outcome` is for in-memory results of handlers,
    which is then additionally converted before being storing as a state.

    Active handler states are those used in .done/.delays for the current
    handling cycle & the current cause. Passive handler states are those
    carried over for logging of counts/extras, and for final state purging,
    but not participating in the current handling cycle.
    """
    started: Optional[datetime.datetime] = None  # None means this information was lost.
    stopped: Optional[datetime.datetime] = None  # None means it is still running (e.g. delayed).
    delayed: Optional[datetime.datetime] = None  # None means it is finished (succeeded/failed).
    retries: int = 0
    success: bool = False
    failure: bool = False

    @property
    def finished(self) -> bool:
        return bool(self.success or self.failure)

    @property
    def sleeping(self) -> bool:
        ts = self.delayed
        now = datetime.datetime.now(datetime.timezone.utc)
        return not self.finished and ts is not None and ts > now

    @property
    def awakened(self) -> bool:
        return bool(not self.finished and not self.sleeping)

    @property
    def runtime(self) -> datetime.timedelta:
        now = datetime.datetime.now(datetime.timezone.utc)
        return now - (self.started if self.started else now)


class State(Mapping[ids.HandlerId, HandlerState]):
    pass


@dataclasses.dataclass
class Cause(invocation.Kwargable):
    """ Base non-specific cause as used in the framework's reactor. """
    logger: typedefs.Logger

    @property
    def _kwargs(self) -> Mapping[str, Any]:
        # Similar to `dataclasses.asdict()`, but not recursive for other dataclasses.
        return {field.name: getattr(self, field.name) for field in dataclasses.fields(self)}


CauseT = TypeVar('CauseT', bound=Cause)


@dataclasses.dataclass(frozen=True)
class Handler:
    """ A handler is a function bound with its behavioral constraints. """
    id: ids.HandlerId
    fn: invocation.Invokable
    param: Optional[Any]
    errors: Optional[ErrorsMode]
    timeout: Optional[float]
    retries: Optional[int]
    backoff: Optional[float]

    # Used in the logs. Overridden in some (but not all) handler types for better log messages.
    def __str__(self) -> str:
        return f"Handler {self.id!r}"

    # Overridden in handlers with fields for causes with field-specific old/new/diff.
    def adjust_cause(self, cause: CauseT) -> CauseT:
        return cause


class LifeCycleFn(Protocol):
    """ A callback type for handlers selection based on the event/cause. """
    def __call__(
            self,
            handlers: Sequence[Handler],
            *,
            state: State,
            **kwargs: Any,
    ) -> Sequence[Handler]: ...


# The task-local context; propagated down the stack instead of multiple kwargs.
# Used in `@kopf.subhandler` and `kopf.execute()` to add/get the sub-handlers.
sublifecycle_var: ContextVar[Optional[LifeCycleFn]] = ContextVar('sublifecycle_var')
subsettings_var: ContextVar[configuration.OperatorSettings] = ContextVar('subsettings_var')
subrefs_var: ContextVar[Iterable[Set[ids.HandlerId]]] = ContextVar('subrefs_var')
handler_var: ContextVar[Handler] = ContextVar('handler_var')
cause_var: ContextVar[Cause] = ContextVar('cause_var')


ExtraContext = Callable[[], AsyncContextManager[None]]


@contextlib.asynccontextmanager
async def no_extra_context() -> AsyncIterator[None]:
    yield


async def execute_handlers_once(
        lifecycle: LifeCycleFn,
        settings: configuration.OperatorSettings,
        handlers: Collection[Handler],
        cause: Cause,
        state: State,
        extra_context: ExtraContext = no_extra_context,
        default_errors: ErrorsMode = ErrorsMode.TEMPORARY,
) -> Mapping[ids.HandlerId, Outcome]:
    """
    Call the next handler(s) from the chain of the handlers.

    Keep the record on the progression of the handlers in the object's state,
    and use it on the next invocation to determined which handler(s) to call.

    This routine is used both for the global handlers (via global registry),
    and for the sub-handlers (via a simple registry of the current handler).
    """

    # Filter and select the handlers to be executed right now, on this event reaction cycle.
    handlers_todo = [h for h in handlers if state[h.id].awakened]
    handlers_plan = lifecycle(handlers_todo, state=state, **cause.kwargs)

    # Execute all planned (selected) handlers in one event reaction cycle, even if there are a few.
    outcomes: MutableMapping[ids.HandlerId, Outcome] = {}
    for handler in handlers_plan:
        outcome = await execute_handler_once(
            settings=settings,
            handler=handler,
            state=state[handler.id],
            cause=cause,
            lifecycle=lifecycle,  # just a default for the sub-handlers, not used directly.
            extra_context=extra_context,
            default_errors=default_errors,
        )
        outcomes[handler.id] = outcome

    return outcomes


async def execute_handler_once(
        settings: configuration.OperatorSettings,
        handler: Handler,
        cause: Cause,
        state: HandlerState,
        lifecycle: Optional[LifeCycleFn] = None,
        extra_context: ExtraContext = no_extra_context,
        default_errors: ErrorsMode = ErrorsMode.TEMPORARY,
) -> Outcome:
    """
    Execute one and only one handler for one and only one time.

    *Execution* means not just *calling* the handler in properly set context
    (see `_call_handler`), but also interpreting its result and errors, and
    wrapping them into am `Outcome` object -- to be stored in the state.

    The *execution* can be long -- depending on how the handler is implemented.
    For daemons, it is normal to run for hours and days if needed.
    This is different from the regular handlers, which are supposed
    to be finished as soon as possible.

    This method is not supposed to raise any exceptions from the handlers:
    exceptions mean the failure of execution itself.
    """
    errors_mode = handler.errors if handler.errors is not None else default_errors
    backoff = handler.backoff if handler.backoff is not None else DEFAULT_RETRY_DELAY
    logger = cause.logger

    # Mutable accumulator for all the sub-handlers of any level deep; populated in `kopf.execute`.
    subrefs: Set[ids.HandlerId] = set()

    # The exceptions are handled locally and are not re-raised, to keep the operator running.
    try:
        logger.debug(f"{handler} is invoked.")

        if handler.timeout is not None and state.runtime.total_seconds() >= handler.timeout:
            raise HandlerTimeoutError(f"{handler} has timed out after {state.runtime}.")

        if handler.retries is not None and state.retries >= handler.retries:
            raise HandlerRetriesError(f"{handler} has exceeded {state.retries} retries.")

        result = await invoke_handler(
            handler=handler,
            cause=cause,
            retry=state.retries,
            started=state.started or datetime.datetime.now(datetime.timezone.utc),  # "or" is for type-checking.
            runtime=state.runtime,
            settings=settings,
            lifecycle=lifecycle,  # just a default for the sub-handlers, not used directly.
            extra_context=extra_context,
            subrefs=subrefs,
        )

    # The cancellations are an excepted way of stopping the handler. Especially for daemons.
    except asyncio.CancelledError:
        logger.warning(f"{handler} is cancelled. Will escalate.")
        raise

    # Unfinished children cause the regular retry, but with less logging and event reporting.
    except HandlerChildrenRetry as e:
        logger.debug(f"{handler} has unfinished sub-handlers. Will retry soon.")
        return Outcome(final=False, exception=e, delay=e.delay, subrefs=subrefs)

    # Definitely a temporary error, regardless of the error strictness.
    except TemporaryError as e:
        logger.error(f"{handler} failed temporarily: {str(e) or repr(e)}")
        return Outcome(final=False, exception=e, delay=e.delay, subrefs=subrefs)

    # Same as permanent errors below, but with better logging for our internal cases.
    except HandlerTimeoutError as e:
        logger.error(f"{str(e) or repr(e)}")  # already formatted
        return Outcome(final=True, exception=e, subrefs=subrefs)
        # TODO: report the handling failure somehow (beside logs/events). persistent status?

    # Definitely a permanent error, regardless of the error strictness.
    except PermanentError as e:
        logger.error(f"{handler} failed permanently: {str(e) or repr(e)}")
        return Outcome(final=True, exception=e, subrefs=subrefs)
        # TODO: report the handling failure somehow (beside logs/events). persistent status?

    # Regular errors behave as either temporary or permanent depending on the error strictness.
    except Exception as e:
        if errors_mode == ErrorsMode.IGNORED:
            logger.exception(f"{handler} failed with an exception. Will ignore.")
            return Outcome(final=True, subrefs=subrefs)
        elif errors_mode == ErrorsMode.TEMPORARY:
            logger.exception(f"{handler} failed with an exception. Will retry.")
            return Outcome(final=False, exception=e, delay=backoff, subrefs=subrefs)
        elif errors_mode == ErrorsMode.PERMANENT:
            logger.exception(f"{handler} failed with an exception. Will stop.")
            return Outcome(final=True, exception=e, subrefs=subrefs)
            # TODO: report the handling failure somehow (beside logs/events). persistent status?
        else:
            raise RuntimeError(f"Unknown mode for errors: {errors_mode!r}")

    # No errors means the handler should be excluded from future runs in this reaction cycle.
    else:
        logger.info(f"{handler} succeeded.")
        return Outcome(final=True, result=result, subrefs=subrefs)


async def invoke_handler(
        *,
        handler: Handler,
        cause: Cause,
        retry: int,
        started: datetime.datetime,
        runtime: datetime.timedelta,
        settings: configuration.OperatorSettings,
        lifecycle: Optional[LifeCycleFn],
        subrefs: Set[ids.HandlerId],
        extra_context: ExtraContext,
) -> Optional[Result]:
    """
    Invoke one handler only, according to the calling conventions.

    Specifically, calculate the handler-specific fields (e.g. field diffs).

    Ensure the global context for this asyncio task is set to the handler and
    its cause -- for proper population of the sub-handlers via the decorators
    (see `@kopf.subhandler`).
    """

    # For the field-handlers, the old/new/diff values must match the field, not the whole object.
    cause = handler.adjust_cause(cause)

    # The context makes it possible and easy to pass the kwargs _through_ the user-space handlers:
    # from the framework to the framework's helper functions (e.g. sub-handling, hierarchies, etc).
    with invocation.context([
        (sublifecycle_var, lifecycle),
        (subsettings_var, settings),
        (subrefs_var, list(subrefs_var.get([])) + [subrefs]),
        (handler_var, handler),
        (cause_var, cause),
    ]):
        async with extra_context():
            result = await invocation.invoke(
                handler.fn,
                settings=settings,
                kwargsrc=cause,
                kwargs=dict(
                    param=handler.param,
                    retry=retry,
                    started=started,
                    runtime=runtime,
                ),
            )

            # Since we know that we invoked the handler, we cast "any" result to a handler result.
            return Result(result)



"""
All the structures needed for Kubernetes patching.

Currently, it is implemented via a JSON merge-patch (RFC 7386),
i.e. a simple dictionary with field overrides, and ``None`` for field deletions.

In the future, it can be extended to a standalone object, which exposes
a dict-like behaviour, and remembers the changes in order of their execution,
and then generates the JSON patch (RFC 6902).
"""
import collections.abc
from typing import Any, Dict, List, MutableMapping, Optional

from typing_extensions import Literal, TypedDict

from kopf._cogs.structs import bodies, dicts

JSONPatchOp = Literal["add", "replace", "remove"]


def _escaped_path(keys: List[str]) -> str:
    """Provides an appropriately escaped path for JSON Patches.

    See https://datatracker.ietf.org/doc/html/rfc6901#section-3 for more details.
    """
    return '/'.join(map(lambda key: key.replace('~', '~0').replace('/', '~1'), keys))


class JSONPatchItem(TypedDict, total=False):
    op: JSONPatchOp
    path: str
    value: Optional[Any]


JSONPatch = List[JSONPatchItem]


class MetaPatch(dicts.MutableMappingView[str, Any]):
    _labels: dicts.MutableMappingView[str, Optional[str]]
    _annotations: dicts.MutableMappingView[str, Optional[str]]

    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'metadata')
        self._labels = dicts.MutableMappingView(self, 'labels')
        self._annotations = dicts.MutableMappingView(self, 'annotations')

    @property
    def labels(self) -> dicts.MutableMappingView[str, Optional[str]]:
        return self._labels

    @property
    def annotations(self) -> dicts.MutableMappingView[str, Optional[str]]:
        return self._annotations


class SpecPatch(dicts.MutableMappingView[str, Any]):
    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'spec')


class StatusPatch(dicts.MutableMappingView[str, Any]):
    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'status')


# Event-handling structures, used internally in the framework and handlers only.
class Patch(Dict[str, Any]):

    def __init__(
        self,
        __src: Optional[MutableMapping[str, Any]] = None,
        body: Optional[bodies.RawBody] = None
    ) -> None:
        super().__init__(__src or {})
        self._meta = MetaPatch(self)
        self._spec = SpecPatch(self)
        self._status = StatusPatch(self)
        self._original = body

    @property
    def metadata(self) -> MetaPatch:
        return self._meta

    @property
    def meta(self) -> MetaPatch:
        return self._meta

    @property
    def spec(self) -> SpecPatch:
        return self._spec

    @property
    def status(self) -> StatusPatch:
        return self._status

    def as_json_patch(self) -> JSONPatch:
        return [] if not self else self._as_json_patch(self, keys=[''])

    def _as_json_patch(self, value: object, keys: List[str]) -> JSONPatch:
        result: JSONPatch = []
        if value is None:
            result.append(JSONPatchItem(op='remove', path=_escaped_path(keys)))
        elif len(keys) > 1 and self._original and not self._is_in_original_path(keys):
            result.append(JSONPatchItem(op='add', path=_escaped_path(keys), value=value))
        elif isinstance(value, collections.abc.Mapping) and value:
            for key, val in value.items():
                result.extend(self._as_json_patch(val, keys + [key]))
        else:
            result.append(JSONPatchItem(op='replace', path=_escaped_path(keys), value=value))
        return result

    def _is_in_original_path(self, keys: List[str]) -> bool:
        _search = self._original
        for key in keys:
            if key == '':
                continue
            try:
                _search = _search[key]  # type: ignore
            except (KeyError, TypeError):
                return False
        return True



import logging

import kopf
from kopf._core.actions.execution import ErrorsMode, PermanentError, TemporaryError


async def test_timer_stopped_on_permanent_error(
        settings, resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle):
    caplog.set_level(logging.DEBUG)

    @kopf.timer(*resource, id='fn', backoff=0.01, interval=1.0)
    async def fn(**kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()
        kwargs['stopped']._setter.set()  # to exit the cycle
        raise PermanentError("boo!")

    event_object = {'metadata': {'finalizers': [settings.persistence.finalizer]}}
    await simulate_cycle(event_object)

    await dummy.steps['called'].wait()
    await dummy.wait_for_daemon_done()

    assert dummy.mock.call_count == 1
    assert k8s_mocked.sleep.call_count == 1  # one for each retry
    assert k8s_mocked.sleep.call_args_list[0][0][0] == 1.0

    assert_logs([
        "Timer 'fn' failed permanently: boo!",
    ], prohibited=[
        "Timer 'fn' succeeded.",
    ])


async def test_timer_stopped_on_arbitrary_errors_with_mode_permanent(
        settings, resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle):
    caplog.set_level(logging.DEBUG)

    @kopf.timer(*resource, id='fn', errors=ErrorsMode.PERMANENT, backoff=0.01, interval=1.0)
    async def fn(**kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()
        kwargs['stopped']._setter.set()  # to exit the cycle
        raise Exception("boo!")

    event_object = {'metadata': {'finalizers': [settings.persistence.finalizer]}}
    await simulate_cycle(event_object)

    await dummy.steps['called'].wait()
    await dummy.wait_for_daemon_done()

    assert dummy.mock.call_count == 1
    assert k8s_mocked.sleep.call_count == 1  # one for each retry
    assert k8s_mocked.sleep.call_args_list[0][0][0] == 1.0

    assert_logs([
        "Timer 'fn' failed with an exception. Will stop.",
    ], prohibited=[
        "Timer 'fn' succeeded.",
    ])


async def test_timer_retried_on_temporary_error(
        settings, resource, dummy, manual_time,
        caplog, assert_logs, k8s_mocked, simulate_cycle):
    caplog.set_level(logging.DEBUG)

    @kopf.timer(*resource, id='fn', backoff=1.0, interval=1.0)
    async def fn(retry, **kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()
        if not retry:
            raise TemporaryError("boo!", delay=1.0)
        else:
            kwargs['stopped']._setter.set()  # to exit the cycle
            dummy.steps['finish'].set()

    event_object = {'metadata': {'finalizers': [settings.persistence.finalizer]}}
    await simulate_cycle(event_object)

    await dummy.steps['called'].wait()
    await dummy.steps['finish'].wait()
    await dummy.wait_for_daemon_done()

    assert k8s_mocked.sleep.call_count == 2  # one for each retry
    assert k8s_mocked.sleep.call_args_list[0][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[1][0][0] == 1.0  # interval

    assert_logs([
        "Timer 'fn' failed temporarily: boo!",
        "Timer 'fn' succeeded.",
    ])


async def test_timer_retried_on_arbitrary_error_with_mode_temporary(
        settings, resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle, manual_time):
    caplog.set_level(logging.DEBUG)

    @kopf.timer(*resource, id='fn', errors=ErrorsMode.TEMPORARY, backoff=1.0, interval=1.0)
    async def fn(retry, **kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()
        if not retry:
            raise Exception("boo!")
        else:
            kwargs['stopped']._setter.set()  # to exit the cycle
            dummy.steps['finish'].set()

    event_object = {'metadata': {'finalizers': [settings.persistence.finalizer]}}
    await simulate_cycle(event_object)

    await dummy.steps['called'].wait()
    await dummy.steps['finish'].wait()
    await dummy.wait_for_daemon_done()

    assert k8s_mocked.sleep.call_count == 2  # one for each retry
    assert k8s_mocked.sleep.call_args_list[0][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[1][0][0] == 1.0  # interval

    assert_logs([
        "Timer 'fn' failed with an exception. Will retry.",
        "Timer 'fn' succeeded.",
    ])


async def test_timer_retried_until_retries_limit(
        resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle, manual_time):
    caplog.set_level(logging.DEBUG)

    @kopf.timer(*resource, id='fn', retries=3, interval=1.0)
    async def fn(**kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()
        if dummy.mock.call_count >= 5:
            kwargs['stopped']._setter.set()  # to exit the cycle
        raise TemporaryError("boo!", delay=1.0)

    await simulate_cycle({})
    await dummy.steps['called'].wait()
    await dummy.wait_for_daemon_done()

    assert k8s_mocked.sleep.call_count >= 4  # one for each retry
    assert k8s_mocked.sleep.call_args_list[0][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[1][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[2][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[3][0][0] == 1.0  # interval


async def test_timer_retried_until_timeout(
        resource, dummy, caplog, assert_logs, k8s_mocked, simulate_cycle, manual_time):
    caplog.set_level(logging.DEBUG)

    @kopf.timer(*resource, id='fn', timeout=3.0, interval=1.0)
    async def fn(**kwargs):
        dummy.mock()
        dummy.kwargs = kwargs
        dummy.steps['called'].set()
        if dummy.mock.call_count >= 5:
            kwargs['stopped']._setter.set()  # to exit the cycle
        raise TemporaryError("boo!", delay=1.0)

    await simulate_cycle({})
    await dummy.steps['called'].wait()
    await dummy.wait_for_daemon_done()

    assert k8s_mocked.sleep.call_count >= 4  # one for each retry
    assert k8s_mocked.sleep.call_args_list[0][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[1][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[2][0][0] == [1.0]  # delays
    assert k8s_mocked.sleep.call_args_list[3][0][0] == 1.0  # interval



import asyncio

from kopf._cogs.aiokits.aiotasks import wait


async def test_wait_with_no_tasks():
    done, pending = await wait([])
    assert not done
    assert not pending


async def test_wait_with_timeout():
    flag = asyncio.Event()
    task = asyncio.create_task(flag.wait())
    done, pending = await wait([task], timeout=0.01)
    assert not done
    assert pending == {task}
    flag.set()
    await task



import asyncio
import logging

import pytest

from kopf._cogs.aiokits.aiotasks import create_guarded_task, reraise


class Error(Exception):
    pass


async def fail(msg: str) -> None:
    raise Error(msg)


async def delay(n: float) -> None:
    await asyncio.sleep(n)


async def sample() -> None:
    pass


async def test_guard_logs_on_exit(assert_logs, caplog):
    caplog.set_level(0)
    logger = logging.getLogger()
    await create_guarded_task(sample(), name='this task', logger=logger)
    assert_logs(["This task has finished unexpectedly"])


async def test_guard_logs_on_failure(assert_logs, caplog):
    caplog.set_level(0)
    logger = logging.getLogger()
    task = create_guarded_task(coro=fail("boo!"), name='this task', logger=logger)
    await asyncio.wait([task], timeout=0.01)  # let it start & react
    assert_logs(["This task has failed: boo!"])


async def test_guard_logs_on_cancellation(assert_logs, caplog):
    caplog.set_level(0)
    logger = logging.getLogger()
    task = create_guarded_task(coro=delay(1), name='this task', logger=logger)
    await asyncio.wait([task], timeout=0.01)  # let it start
    task.cancel()
    await asyncio.wait([task], timeout=0.01)  # let it react
    assert_logs(["This task is cancelled"])


async def test_guard_is_silent_when_finishable(assert_logs, caplog):
    caplog.set_level(0)
    logger = logging.getLogger()
    await create_guarded_task(sample(), name='this task', logger=logger, finishable=True)
    assert_logs([], prohibited=["This task has finished unexpectedly"])
    assert not caplog.messages


async def test_guard_is_silent_when_cancellable(assert_logs, caplog):
    caplog.set_level(0)
    logger = logging.getLogger()
    task = create_guarded_task(coro=delay(1), name='this task', logger=logger, cancellable=True)
    await asyncio.wait([task], timeout=0.01)  # let it start
    task.cancel()
    await asyncio.wait([task], timeout=0.01)  # let it react
    assert_logs([], prohibited=["This task is cancelled"])
    assert not caplog.messages


@pytest.mark.parametrize('finishable', [True, False])
async def test_guard_escalates_on_failure(finishable):
    task = create_guarded_task(coro=fail("boo!"), name='this task', finishable=finishable)
    await asyncio.wait([task], timeout=0.01)  # let it start & react
    with pytest.raises(Error):
        await task


@pytest.mark.parametrize('cancellable', [True, False])
async def test_guard_escalates_on_cancellation(cancellable):
    task = create_guarded_task(coro=delay(1), name='this task', cancellable=cancellable)
    await asyncio.wait([task], timeout=0.01)  # let it start
    task.cancel()
    await asyncio.wait([task], timeout=0.01)  # let it react
    with pytest.raises(asyncio.CancelledError):
        await task


async def test_guard_waits_for_the_flag():
    flag = asyncio.Event()

    task = create_guarded_task(coro=sample(), name='this task', flag=flag)
    await asyncio.wait([task], timeout=0.01)  # let it start
    assert not task.done()

    flag.set()
    await asyncio.wait([task], timeout=0.01)  # let it react
    assert task.done()


async def test_reraise_escalates_errors():
    task = asyncio.create_task(fail("boo!"))
    await asyncio.wait([task], timeout=0.01)  # let it start & react
    with pytest.raises(Error):
        await reraise([task])


async def test_reraise_skips_cancellations():
    task = asyncio.create_task(asyncio.Event().wait())
    done, pending = await asyncio.wait([task], timeout=0.01)  # let it start
    assert not done
    task.cancel()
    done, pending = await asyncio.wait([task], timeout=0.01)  # let it react
    assert done
    await reraise([task])
